"""Cryptography functions used by BasicWallet."""

import hashlib
import logging
import re
from collections import OrderedDict
from typing import Callable, List, Optional, Sequence, Tuple, Union

import nacl.bindings
import nacl.exceptions
import nacl.utils
from marshmallow import ValidationError

from ..utils.jwe import JweEnvelope, JweRecipient, b64url, from_b64url
from .bbs import (
    BbsException,
    create_bls12381g2_keypair,
    sign_messages_bls12381g2,
    verify_signed_messages_bls12381g2,
)
from .did_method import INDY, SOV, DIDMethod
from .error import WalletError
from .key_type import BLS12381G2, ED25519, KeyType
from .util import b58_to_bytes, b64_to_bytes, bytes_to_b58, random_seed

LOGGER = logging.getLogger(__name__)


def create_keypair(
    key_type: KeyType, seed: Optional[bytes] = None
) -> Tuple[bytes, bytes]:
    """Create a public and private keypair from a seed value.

    Args:
        key_type: The type of key to generate
        seed: Seed for keypair

    Raises:
        WalletError: If the key type is not supported

    Returns:
        A tuple of (public key, secret key)

    """
    if key_type == ED25519:
        return create_ed25519_keypair(seed)
    elif key_type == BLS12381G2:
        # This ensures python won't crash if bbs is not installed and not used

        return create_bls12381g2_keypair(seed)
    else:
        raise WalletError(f"Unsupported key type: {key_type.key_type}")


def create_ed25519_keypair(seed: Optional[bytes] = None) -> Tuple[bytes, bytes]:
    """Create a public and private ed25519 keypair from a seed value.

    Args:
        seed: Seed for keypair

    Returns:
        A tuple of (public key, secret key)

    """
    if not seed:
        seed = random_seed()
    pk, sk = nacl.bindings.crypto_sign_seed_keypair(seed)
    return pk, sk


def seed_to_did(seed: str, method: Optional[DIDMethod] = SOV) -> str:
    """Derive a DID from a seed value.

    Args:
        seed: The seed to derive
        method: The DID method to use

    Returns:
        The DID derived from the seed

    """
    seed = validate_seed(seed)
    verkey, _ = create_ed25519_keypair(seed)
    if method == SOV:
        return bytes_to_b58(verkey[:16])
    if method == INDY:
        # Hash the verkey, take the first 16 bytes, and convert to a base58 string
        hashed_verkey = hashlib.sha256(verkey).digest()
        did = bytes_to_b58(hashed_verkey[:16])
        return f"did:indy:{did}"
    raise WalletError(f"Unsupported DID method: {method.method_name}")


def did_is_self_certified(did: str, verkey: str) -> bool:
    """Check if the DID is self certified.

    Args:
        did: DID string
        verkey: VERKEY string
    """
    ABBREVIATED_VERKEY_REGEX = "^~[1-9A-HJ-NP-Za-km-z]{21,22}$"
    if re.search(ABBREVIATED_VERKEY_REGEX, verkey):
        return True
    verkey_bytes = b58_to_bytes(verkey)
    did_from_verkey = bytes_to_b58(verkey_bytes[:16])
    if did == did_from_verkey:
        return True
    return False


def sign_pk_from_sk(secret: bytes) -> bytes:
    """Extract the verkey from a secret signing key."""
    seed_len = nacl.bindings.crypto_sign_SEEDBYTES
    return secret[seed_len:]


def validate_seed(seed: Union[str, bytes, None]) -> bytes:
    """Convert a seed parameter to standard format and check length.

    Args:
        seed: The seed to validate

    Returns:
        The validated and encoded seed

    """
    if not seed:
        return None
    if isinstance(seed, str):
        if "=" in seed:
            seed = b64_to_bytes(seed)
        else:
            seed = seed.encode("ascii")
    if not isinstance(seed, bytes):
        raise WalletError("Seed value is not a string or bytes")
    if len(seed) != 32:
        raise WalletError("Seed value must be 32 bytes in length")
    return seed


def sign_message(
    message: Union[List[bytes], bytes], secret: bytes, key_type: KeyType
) -> bytes:
    """Sign message(s) using a private signing key.

    Args:
        message: The message(s) to sign
        secret: The private signing key
        key_type: The key type to derive the signature algorithm from

    Returns:
        bytes: The signature

    """
    # Make messages list if not already for easier checking going forward
    messages = message if isinstance(message, list) else [message]

    if key_type == ED25519:
        if len(messages) > 1:
            raise WalletError("ed25519 can only sign a single message")

        return sign_message_ed25519(
            message=messages[0],
            secret=secret,
        )
    elif key_type == BLS12381G2:
        return sign_messages_bls12381g2(messages=messages, secret=secret)
    else:
        raise WalletError(f"Unsupported key type: {key_type.key_type}")


def sign_message_ed25519(message: bytes, secret: bytes) -> bytes:
    """Sign message using an ed25519 private signing key.

    This function takes a message and a private signing key as input and returns the
    signature of the message using the ed25519 algorithm.

    Args:
        message (bytes): The message to sign.
        secret (bytes): The private signing key.

    Returns:
        bytes: The signature of the message.

    """
    result = nacl.bindings.crypto_sign(message, secret)
    sig = result[: nacl.bindings.crypto_sign_BYTES]
    return sig


def verify_signed_message(
    message: Union[List[bytes], bytes],
    signature: bytes,
    verkey: bytes,
    key_type: KeyType,
) -> bool:
    """Verify a signed message according to a public verification key.

    Args:
        message: The message(s) to verify
        signature: The signature to verify
        verkey: The verkey to use in verification
        key_type: The key type to derive the signature verification algorithm from

    Returns:
        True if verified, else False

    """
    # Make messages list if not already for easier checking going forward
    messages = message if isinstance(message, list) else [message]

    if key_type == ED25519:
        if len(messages) > 1:
            raise WalletError("ed25519 can only verify a single message")

        return verify_signed_message_ed25519(
            message=messages[0], signature=signature, verkey=verkey
        )
    elif key_type == BLS12381G2:
        try:
            return verify_signed_messages_bls12381g2(
                messages=messages, signature=signature, public_key=verkey
            )
        except BbsException as e:
            raise WalletError("Unable to verify message") from e
    else:
        raise WalletError(f"Unsupported key type: {key_type.key_type}")


def verify_signed_message_ed25519(
    message: bytes, signature: bytes, verkey: bytes
) -> bool:
    """Verify an ed25519 signed message according to a public verification key.

    Args:
        message: The message to verify
        signature: The signature to verify
        verkey: The verkey to use in verification

    Returns:
        True if verified, else False

    """
    try:
        nacl.bindings.crypto_sign_open(signature + message, verkey)
    except nacl.exceptions.BadSignatureError:
        return False
    return True


def add_pack_recipients(
    wrapper: JweEnvelope,
    cek: bytes,
    to_verkeys: Sequence[bytes],
    from_secret: Optional[bytes] = None,
):
    """Assemble the recipients block of a packed message.

    Args:
        wrapper: The envelope to add recipients to
        cek: The content encryption key
        to_verkeys: Verkeys of recipients
        from_secret: Secret to use for signing keys

    Returns:
        A tuple of (json result, key)

    """
    for target_vk in to_verkeys:
        target_pk = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(target_vk)
        if from_secret:
            sender_pk = sign_pk_from_sk(from_secret)
            sender_vk = bytes_to_b58(sender_pk).encode("utf-8")
            enc_sender = nacl.bindings.crypto_box_seal(sender_vk, target_pk)
            sk = nacl.bindings.crypto_sign_ed25519_sk_to_curve25519(from_secret)

            nonce = nacl.utils.random(nacl.bindings.crypto_box_NONCEBYTES)
            enc_cek = nacl.bindings.crypto_box(cek, nonce, target_pk, sk)
            wrapper.add_recipient(
                JweRecipient(
                    encrypted_key=enc_cek,
                    header=OrderedDict(
                        [
                            ("kid", bytes_to_b58(target_vk)),
                            ("sender", b64url(enc_sender)),
                            ("iv", b64url(nonce)),
                        ]
                    ),
                )
            )
        else:
            enc_sender = None
            nonce = None
            enc_cek = nacl.bindings.crypto_box_seal(cek, target_pk)
            wrapper.add_recipient(
                JweRecipient(
                    encrypted_key=enc_cek, header={"kid": bytes_to_b58(target_vk)}
                )
            )


def ed25519_pk_to_curve25519(public_key: bytes) -> bytes:
    """Covert a public Ed25519 key to a public Curve25519 key as bytes."""
    return nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(public_key)


def encrypt_plaintext(
    message: str, add_data: bytes, key: bytes
) -> Tuple[bytes, bytes, bytes]:
    """Encrypt the payload of a packed message.

    Args:
        message (str): The message to encrypt.
        add_data (bytes): Additional data to include in the encryption.
        key (bytes): The key used for encryption.

    Returns:
        Tuple[bytes, bytes, bytes]: A tuple containing the ciphertext, nonce, and tag.

    """
    nonce = nacl.utils.random(nacl.bindings.crypto_aead_chacha20poly1305_ietf_NPUBBYTES)
    message_bin = message.encode("utf-8")
    output = nacl.bindings.crypto_aead_chacha20poly1305_ietf_encrypt(
        message_bin, add_data, nonce, key
    )
    mlen = len(message)
    ciphertext = output[:mlen]
    tag = output[mlen:]
    return ciphertext, nonce, tag


def decrypt_plaintext(
    ciphertext: bytes, recips_bin: bytes, nonce: bytes, key: bytes
) -> str:
    """Decrypt the payload of a packed message.

    Args:
        ciphertext (bytes): The encrypted payload to be decrypted.
        recips_bin (bytes): The binary representation of the recipients' public keys.
        nonce (bytes): The nonce used for encryption.
        key (bytes): The secret key used for encryption.

    Returns:
        str: The decrypted string.

    """
    output = nacl.bindings.crypto_aead_chacha20poly1305_ietf_decrypt(
        ciphertext, recips_bin, nonce, key
    )
    return output.decode("utf-8")


def encode_pack_message(
    message: str, to_verkeys: Sequence[bytes], from_secret: Optional[bytes] = None
) -> bytes:
    """Assemble a packed message for a set of recipients, optionally including the sender.

    Args:
        message: The message to pack
        to_verkeys: The verkeys to pack the message for
        from_secret: The sender secret

    Returns:
        The encoded message

    """
    wrapper = JweEnvelope(with_protected_recipients=True, with_flatten_recipients=False)
    cek = nacl.bindings.crypto_secretstream_xchacha20poly1305_keygen()
    add_pack_recipients(wrapper, cek, to_verkeys, from_secret)
    wrapper.set_protected(
        OrderedDict(
            [
                ("enc", "xchacha20poly1305_ietf"),
                ("typ", "JWM/1.0"),
                ("alg", "Authcrypt" if from_secret else "Anoncrypt"),
            ]
        ),
    )
    ciphertext, nonce, tag = encrypt_plaintext(message, wrapper.protected_bytes, cek)
    wrapper.set_payload(ciphertext, nonce, tag)
    return wrapper.to_json().encode("utf-8")


def decode_pack_message(
    enc_message: bytes, find_key: Callable
) -> Tuple[str, Optional[str], str]:
    """Decode a packed message.

    Disassemble and unencrypt a packed message, returning the message content,
    verification key of the sender (if available), and verification key of the
    recipient.

    Args:
        enc_message: The encrypted message
        find_key: Function to retrieve private key

    Returns:
        A tuple of (message, sender_vk, recip_vk)

    Raises:
        ValueError: If the packed message is invalid
        ValueError: If the packed message recipients are invalid
        ValueError: If the pack algorithm is unsupported
        ValueError: If the sender's public key was not provided

    """
    wrapper, recips, is_authcrypt = decode_pack_message_outer(enc_message)
    payload_key, sender_vk = None, None
    for recip_vk in recips:
        recip_secret = find_key(recip_vk)
        if recip_secret:
            payload_key, sender_vk = extract_payload_key(recips[recip_vk], recip_secret)
            break

    if not payload_key:
        raise ValueError(
            "No corresponding recipient key found in {}".format(tuple(recips))
        )
    if not sender_vk and is_authcrypt:
        raise ValueError("Sender public key not provided for Authcrypt message")

    message = decode_pack_message_payload(wrapper, payload_key)
    return message, sender_vk, recip_vk


def decode_pack_message_outer(enc_message: bytes) -> Tuple[dict, dict, bool]:
    """Decode the outer wrapper of a packed message and extract the recipients.

    Args:
        enc_message: The encrypted message

    Returns: a tuple of the decoded wrapper, recipients, and authcrypt flag

    """
    try:
        wrapper = JweEnvelope.from_json(enc_message)
    except ValidationError as err:
        LOGGER.error(err)
        raise ValueError("Invalid packed message")

    alg = wrapper.protected.get("alg")
    is_authcrypt = alg == "Authcrypt"
    if not is_authcrypt and alg != "Anoncrypt":
        raise ValueError("Unsupported pack algorithm: {}".format(alg))

    recips = extract_pack_recipients(wrapper.recipients)
    return wrapper, recips, is_authcrypt


def decode_pack_message_payload(wrapper: JweEnvelope, payload_key: bytes) -> str:
    """Decode the payload of a packed message once the CEK is known.

    Args:
        wrapper: The decoded message wrapper
        payload_key: The decrypted payload key

    """
    payload_bin = wrapper.ciphertext + wrapper.tag
    message = decrypt_plaintext(
        payload_bin, wrapper.protected_bytes, wrapper.iv, payload_key
    )
    return message


def extract_pack_recipients(recipients: Sequence[JweRecipient]) -> dict:
    """Extract the pack message recipients into a dict indexed by verkey.

    Args:
        recipients: Recipients to locate

    Raises:
        ValueError: If the recipients block is mal-formatted

    """
    result = {}
    for recip in recipients:
        recip_vk_b58 = recip.header.get("kid")
        if not recip_vk_b58:
            raise ValueError("Blank recipient key")
        if recip_vk_b58 in result:
            raise ValueError("Duplicate recipient key")

        sender_b64 = recip.header.get("sender")
        enc_sender = from_b64url(sender_b64) if sender_b64 else None

        nonce_b64 = recip.header.get("iv")
        if sender_b64 and not nonce_b64:
            raise ValueError("Missing iv")
        elif not sender_b64 and nonce_b64:
            raise ValueError("Unexpected iv")
        nonce = from_b64url(nonce_b64) if nonce_b64 else None

        result[recip_vk_b58] = {
            "sender": enc_sender,
            "nonce": nonce,
            "key": recip.encrypted_key,
        }
    return result


def extract_payload_key(sender_cek: dict, recip_secret: bytes) -> Tuple[bytes, str]:
    """Extract the payload key from pack recipient details.

    Returns: A tuple of the CEK and sender verkey
    """
    recip_vk = sign_pk_from_sk(recip_secret)
    recip_pk = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(recip_vk)
    recip_sk = nacl.bindings.crypto_sign_ed25519_sk_to_curve25519(recip_secret)

    if sender_cek["nonce"] and sender_cek["sender"]:
        sender_vk_bin = nacl.bindings.crypto_box_seal_open(
            sender_cek["sender"], recip_pk, recip_sk
        )
        sender_vk = sender_vk_bin.decode("utf-8")
        sender_pk = nacl.bindings.crypto_sign_ed25519_pk_to_curve25519(
            b58_to_bytes(sender_vk_bin)
        )
        cek = nacl.bindings.crypto_box_open(
            sender_cek["key"], sender_cek["nonce"], sender_pk, recip_sk
        )
    else:
        sender_vk = None
        cek = nacl.bindings.crypto_box_seal_open(sender_cek["key"], recip_pk, recip_sk)
    return cek, sender_vk



"""Multitenant error classes."""

from ..core.error import BaseError


class MultitenantManagerError(BaseError):
    """Generic multitenant error."""


class InvalidTokenError(MultitenantManagerError):
    """Exception raised for invalid tokens."""

    def __init__(self, message: str = "Token not valid"):
        """Initialize an instance of InvalidTokenError."""
        super().__init__(message)


class MissingProfileError(MultitenantManagerError):
    """Exception raised when a profile is missing."""

    def __init__(self, message: str = "Missing profile"):
        """Initialize an instance of MissingProfileError."""
        super().__init__(message)


class WalletAlreadyExistsError(MultitenantManagerError):
    """Exception raised when a wallet already exists."""

    def __init__(self, wallet_name: str):
        """Initialize an instance of WalletAlreadyExistsError."""
        message = f"Wallet with name {wallet_name} already exists"
        super().__init__(message)


class WalletKeyMissingError(MultitenantManagerError):
    """Wallet key missing exception."""

    def __init__(self, message: str = "Missing key to open wallet"):
        """Initialize an instance of WalletKeyMissingError."""
        super().__init__(message)



from typing import Optional
from unittest import TestCase

from marshmallow import EXCLUDE, fields

from ...models.base import BaseModel, BaseModelSchema
from ..base import BaseDecoratorSet
from ..default import DEFAULT_MODELS, DecoratorSet


class SimpleModel(BaseModel):
    class Meta:
        schema_class = "SimpleModelSchema"

    def __init__(
        self,
        *,
        value: Optional[str] = None,
        handled_decorator: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.handled_decorator = handled_decorator
        self.value = value


class SimpleModelSchema(BaseModelSchema):
    class Meta:
        model_class = SimpleModel
        unknown = EXCLUDE

    value = fields.Str(required=True)
    handled_decorator = fields.Str(required=False, data_key="handled~decorator")


class TestDecoratorSet(TestCase):
    def test_deco_set(self):
        deco_set = DecoratorSet()
        assert all(k in deco_set.models for k in DEFAULT_MODELS)

    def test_extract(self):
        decor_value = {}
        message = {"~decorator": decor_value, "one": "TWO"}

        decors = BaseDecoratorSet()
        remain = decors.extract_decorators(message)

        # check original is unmodified
        assert "~decorator" in message

        assert decors["decorator"] is decor_value
        assert remain == {"one": "TWO"}

    def test_dict(self):
        decors = BaseDecoratorSet()
        decors["test"] = "TEST"
        assert decors["test"] == "TEST"
        result = decors.to_dict()
        assert result == {"~test": "TEST"}

    def test_decorator_model(self):
        decor_value = {}
        message = {"~test": {"value": "TEST"}}

        decors = BaseDecoratorSet()
        decors.add_model("test", SimpleModel)
        remain = decors.extract_decorators(message, SimpleModelSchema)

        tested = decors["test"]
        assert isinstance(tested, SimpleModel) and tested.value == "TEST"

        result = decors.to_dict()
        assert result == message

    def test_field_decorator(self):
        decor_value = {}
        message = {"test~decorator": decor_value, "one": "TWO"}

        decors = BaseDecoratorSet()
        remain = decors.extract_decorators(message, SimpleModelSchema)

        # check original is unmodified
        assert "test~decorator" in message

        assert decors.field("test")
        assert decors.field("test")["decorator"] is decor_value
        assert remain == {"one": "TWO"}
        assert "test~decorator" in decors.to_dict()

    def test_skip_decorator(self):
        decor_value = {}
        message = {"handled~decorator": decor_value, "one": "TWO"}

        decors = BaseDecoratorSet()
        remain = decors.extract_decorators(message, SimpleModelSchema)

        # check original is unmodified
        assert "handled~decorator" in message

        assert not decors.field("handled")
        assert remain == message
        assert not decors.to_dict()



"""Assertion proof purpose class."""

from datetime import datetime, timedelta
from typing import Optional

from .controller_proof_purpose import ControllerProofPurpose


class AssertionProofPurpose(ControllerProofPurpose):
    """Assertion proof purpose class."""

    term = "assertionMethod"

    def __init__(
        self,
        *,
        date: Optional[datetime] = None,
        max_timestamp_delta: Optional[timedelta] = None,
    ):
        """Initialize new instance of AssertionProofPurpose."""
        super().__init__(
            term=AssertionProofPurpose.term,
            date=date,
            max_timestamp_delta=max_timestamp_delta,
        )



import os
import tempfile

from aiohttp import web
from aiohttp.test_utils import AioHTTPTestCase

from ..http import FetchError, PutError, fetch, fetch_stream, put_file


class TempFile:
    def __init__(self):
        self.name = None

    def __enter__(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.write(b"test")
        file.close()
        self.name = file.name
        return self.name

    def __exit__(self, *args):
        if self.name:
            os.unlink(self.name)


class TestTransportUtils(AioHTTPTestCase):
    async def setUpAsync(self):
        self.fail_calls = 0
        self.succeed_calls = 0
        self.redirects = 0
        await super().setUpAsync()

    async def get_application(self):
        app = web.Application()
        app.add_routes(
            [
                web.get("/fail", self.fail_route),
                web.get("/succeed", self.succeed_route),
                web.put("/fail", self.fail_route),
                web.put("/succeed", self.succeed_route),
                web.put("/redirect", self.redirect_route),
            ]
        )
        return app

    async def fail_route(self, request):
        self.fail_calls += 1
        # avoid aiohttp test server issue: https://github.com/aio-libs/aiohttp/issues/3968
        await request.read()
        raise web.HTTPForbidden()

    async def succeed_route(self, request):
        self.succeed_calls += 1
        ret = web.json_response([True])
        return ret

    async def redirect_route(self, request):
        if self.redirects > 0:
            self.redirects -= 1
            # avoid aiohttp test server issue: https://github.com/aio-libs/aiohttp/issues/3968
            await request.read()
            raise web.HTTPRedirection(f"http://localhost:{self.server.port}/success")
        return await self.succeed_route(request)

    async def test_fetch_stream(self):
        server_addr = f"http://localhost:{self.server.port}"
        stream = await fetch_stream(f"{server_addr}/succeed", session=self.client.session)
        result = await stream.read()
        assert result == b"[true]"
        assert self.succeed_calls == 1

    async def test_fetch_stream_default_client(self):
        server_addr = f"http://localhost:{self.server.port}"
        stream = await fetch_stream(f"{server_addr}/succeed")
        result = await stream.read()
        assert result == b"[true]"
        assert self.succeed_calls == 1

    async def test_fetch_stream_fail(self):
        server_addr = f"http://localhost:{self.server.port}"
        with self.assertRaises(FetchError):
            await fetch_stream(
                f"{server_addr}/fail",
                max_attempts=2,
                interval=0,
                session=self.client.session,
            )
        assert self.fail_calls == 2

    async def test_fetch(self):
        server_addr = f"http://localhost:{self.server.port}"
        result = await fetch(
            f"{server_addr}/succeed", session=self.client.session, json=True
        )
        assert result == [1]
        assert self.succeed_calls == 1

    async def test_fetch_default_client(self):
        server_addr = f"http://localhost:{self.server.port}"
        result = await fetch(f"{server_addr}/succeed", json=True)
        assert result == [1]
        assert self.succeed_calls == 1

    async def test_fetch_fail(self):
        server_addr = f"http://localhost:{self.server.port}"
        with self.assertRaises(FetchError):
            result = await fetch(
                f"{server_addr}/fail",
                max_attempts=2,
                interval=0,
                session=self.client.session,
            )
        assert self.fail_calls == 2

    async def test_put_file_with_session(self):
        server_addr = f"http://localhost:{self.server.port}"
        with TempFile() as tails:
            result = await put_file(
                f"{server_addr}/succeed",
                {"tails": tails},
                {"genesis": "..."},
                session=self.client.session,
                json=True,
            )
        assert result == [True]
        assert self.succeed_calls == 1

    async def test_put_file_default_client(self):
        server_addr = f"http://localhost:{self.server.port}"
        with TempFile() as tails:
            result = await put_file(
                f"{server_addr}/succeed",
                {"tails": tails},
                {"genesis": "..."},
                json=True,
            )
        assert result == [True]
        assert self.succeed_calls == 1

    async def test_put_file_fail(self):
        server_addr = f"http://localhost:{self.server.port}"
        with TempFile() as tails:
            with self.assertRaises(PutError):
                _ = await put_file(
                    f"{server_addr}/fail",
                    {"tails": tails},
                    {"genesis": "..."},
                    max_attempts=2,
                    json=True,
                )
        assert self.fail_calls == 2

    async def test_put_file_redirect(self):
        server_addr = f"http://localhost:{self.server.port}"
        self.redirects = 1
        with TempFile() as tails:
            result = await put_file(
                f"{server_addr}/redirect",
                {"tails": tails},
                {"genesis": "..."},
                max_attempts=2,
                json=True,
            )
        assert result == [True]
        assert self.succeed_calls == 1
        assert self.redirects == 0



from unittest import TestCase

from ...valid import UUID4_EXAMPLE
from ..transport_decorator import TransportDecorator


class TestTransportDecorator(TestCase):
    def test_serialize_load(self):
        deco = TransportDecorator(
            return_route="all",
            return_route_thread=UUID4_EXAMPLE,
            queued_message_count=23,
        )

        assert deco.return_route == "all"
        assert deco.return_route_thread == UUID4_EXAMPLE
        assert deco.queued_message_count == 23

        dumped = deco.serialize()
        loaded = TransportDecorator.deserialize(dumped)



import json
from copy import deepcopy
from time import time
from unittest import IsolatedAsyncioTestCase

import pytest
from anoncreds import W3cCredential
from marshmallow import ValidationError

from .......anoncreds.holder import AnonCredsHolder, AnonCredsHolderError
from .......anoncreds.issuer import AnonCredsIssuer
from .......anoncreds.models.credential_definition import CredDef, GetCredDefResult
from .......anoncreds.models.revocation import GetRevRegDefResult, RevRegDef
from .......anoncreds.registry import AnonCredsRegistry
from .......cache.base import BaseCache
from .......cache.in_memory import InMemoryCache
from .......ledger.base import BaseLedger
from .......ledger.multiple_ledger.ledger_requests_executor import (
    IndyLedgerRequestsExecutor,
)
from .......messaging.credential_definitions.util import CRED_DEF_SENT_RECORD_TYPE
from .......messaging.decorators.attach_decorator import AttachDecorator
from .......multitenant.base import BaseMultitenantManager
from .......multitenant.manager import MultitenantManager
from .......protocols.issue_credential.v2_0.formats.handler import V20CredFormatError
from .......protocols.issue_credential.v2_0.messages.cred_format import V20CredFormat
from .......protocols.issue_credential.v2_0.messages.cred_issue import V20CredIssue
from .......protocols.issue_credential.v2_0.messages.cred_offer import V20CredOffer
from .......protocols.issue_credential.v2_0.messages.cred_proposal import V20CredProposal
from .......protocols.issue_credential.v2_0.messages.cred_request import V20CredRequest
from .......protocols.issue_credential.v2_0.messages.inner.cred_preview import (
    V20CredAttrSpec,
    V20CredPreview,
)
from .......protocols.issue_credential.v2_0.models.cred_ex_record import V20CredExRecord
from .......protocols.issue_credential.v2_0.models.detail.indy import V20CredExRecordIndy
from .......storage.base import BaseStorage
from .......storage.record import StorageRecord
from .......tests import mock
from .......utils.testing import create_test_profile
from .......wallet.askar import AskarWallet
from .......wallet.base import BaseWallet
from .......wallet.did_info import DIDInfo
from ....message_types import (
    ATTACHMENT_FORMAT,
    CRED_20_ISSUE,
    CRED_20_OFFER,
    CRED_20_PROPOSAL,
    CRED_20_REQUEST,
)
from .. import handler as test_module
from ..handler import LOGGER as VCDI_LOGGER
from ..handler import VCDICredFormatHandler

# these are from faber
CRED_PREVIEW_TYPE = "https://didcomm.org/issue-credential/2.0/credential-preview"


TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "bc-reg"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
SCHEMA = {
    "ver": "1.0",
    "id": SCHEMA_ID,
    "name": SCHEMA_NAME,
    "version": "1.0",
    "attrNames": ["legalName", "jurisdictionId", "incorporationDate"],
    "seqNo": SCHEMA_TXN,
}
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:tag1"
CRED_DEF = {
    "ver": "1.0",
    "id": CRED_DEF_ID,
    "schemaId": SCHEMA_TXN,
    "type": "CL",
    "tag": "tag1",
    "value": {
        "primary": {
            "n": "...",
            "s": "...",
            "r": {
                "master_secret": "...",
                "legalName": "...",
                "jurisdictionId": "...",
                "incorporationDate": "...",
            },
            "rctxt": "...",
            "z": "...",
        },
        "revocation": {
            "g": "1 ...",
            "g_dash": "1 ...",
            "h": "1 ...",
            "h0": "1 ...",
            "h1": "1 ...",
            "h2": "1 ...",
            "htilde": "1 ...",
            "h_cap": "1 ...",
            "u": "1 ...",
            "pk": "1 ...",
            "y": "1 ...",
        },
    },
}
REV_REG_DEF_TYPE = "CL_ACCUM"
REV_REG_ID = f"{TEST_DID}:4:{CRED_DEF_ID}:{REV_REG_DEF_TYPE}:tag1"
TAILS_DIR = "/tmp/indy/revocation/tails_files"
TAILS_HASH = "8UW1Sz5cqoUnK9hqQk7nvtKK65t7Chu3ui866J23sFyJ"
TAILS_LOCAL = f"{TAILS_DIR}/{TAILS_HASH}"
REV_REG_DEF = {
    "ver": "1.0",
    "id": REV_REG_ID,
    "revocDefType": "CL_ACCUM",
    "tag": "tag1",
    "credDefId": CRED_DEF_ID,
    "value": {
        "issuanceType": "ISSUANCE_ON_DEMAND",
        "maxCredNum": 5,
        "publicKeys": {"accumKey": {"z": "1 ..."}},
        "tailsHash": TAILS_HASH,
        "tailsLocation": TAILS_LOCAL,
    },
}
VCDI_OFFER = {
    "data_model_versions_supported": ["1.1"],
    "binding_required": True,
    "binding_method": {
        "anoncreds_link_secret": {
            "cred_def_id": CRED_DEF_ID,
            "key_correctness_proof": {
                "c": "123467890",
                "xz_cap": "12345678901234567890",
                "xr_cap": [
                    ["remainder", "1234567890"],
                    ["number", "12345678901234"],
                ],
            },
            "nonce": "803336938981521544311884",
        },
        "didcomm_signed_attachment": {
            "algs_supported": ["EdDSA"],
            "did_methods_supported": ["key"],
            "nonce": "803336938981521544311884",
        },
    },
    "credential": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://w3id.org/security/data-integrity/v2",
            {"@vocab": "https://www.w3.org/ns/credentials/issuer-dependent#"},
        ],
        "type": ["VerifiableCredential"],
        "issuer": "LSTv7AUoTyfqFxZbuAGqKR",
        "credentialSubject": {
            "name": "Alice Smith",
            "date": "2018-05-28",
            "degree": "Maths",
            "birthdate_dateint": "20000331",
            "timestamp": "1711845568",
        },
        "issuanceDate": "2024-01-10T04:44:29.563418Z",
    },
}

VCDI_CRED_REQ = {
    "data_model_version": "2.0",
    "binding_proof": {
        "anoncreds_link_secret": {
            "prover_did": f"did:sov:{TEST_DID}",
            "entropy": f"did:sov:{TEST_DID}",
            "cred_def_id": CRED_DEF_ID,
            "blinded_ms": {
                "u": "10047077609650450290609991930929594521921208780899757965398360086992099381832995073955506958821655372681970112562804577530208651675996528617262693958751195285371230790988741041496869140904046414320278189103736789305088489636024127715978439300785989247215275867951013255925809735479471883338351299180591011255281885961242995409072561940244771612447316409017677474822482928698183528232263803799926211692640155689629903898365777273000566450465466723659861801656618726777274689021162957914736922404694190070274236964163273886807208820068271673047750886130307545831668836096290655823576388755329367886670574352063509727295",
                "ur": "1 10047077609650450290609991930929594521921208780899757965398360086992099381832995073955506958821655372681970112562804577530208651675996528617262693958751195285371230790988741041496869140904046414320278189103736789305088489636024127715978439300785989247215275867951013255925809735479471883338351299180591011255281885961242995409072561940244771612447316409017677474822482928698183528232263803799926211692640155689629903898365777273000566450465466723659861801656618726777274689021162957914736922404694190070274236964163273886807208820068271673047750886130307545831668836096290655823576388755329367886670574352063509727295",
                "hidden_attributes": ["master_secret"],
                "committed_attributes": {},
            },
            "blinded_ms_correctness_proof": {
                "c": "114820252909341277169123380270435575009714169580229908332117664809097619479483",
                "v_dash_cap": "2800797042446023854769298889946111553775800551626595767742635719512900304820391485829151945623333206184503230504182991047567531709613146606620747119977967362375975470346540769137309709302645176745785595101997824808807951935607979085748767054924474772855886854081455495367299835633316236603850924206877781343663290011630380243973434735740532318737134036990657225621660855862337569102753069931768335142276913795486645880476005516655059658346071702100939785144477050087370752056081492070366540039114009106296993876935692142991636251707934248460120048734266848191670576929279282843107501392282445417047087792806945845190270343938754413343820710137866411061071233755924209847337885397612906914410338546708562034035772917684",
                "m_caps": {
                    "master_secret": "27860715812851216521476619601374576486949815748604240358820717458669963808683330226247784299086475972693040811857737248781118170534547319442287676278121026619110648625203644115424"
                },
                "r_caps": {},
            },
            "nonce": "866186615577411311009479",
        },
        "didcomm_signed_attachment": {"attachment_id": "test"},
    },
}

VCDI_CRED = {
    "credential": {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://w3id.org/security/data-integrity/v2",
            {"@vocab": "https://www.w3.org/ns/credentials/issuer-dependent#"},
        ],
        "type": ["VerifiableCredential"],
        "issuer": "LSTv7AUoTyfqFxZbuAGqKR",
        "credentialSubject": {
            "name": "Alice Smith",
            "timestamp": "1711845568",
            "birthdate_dateint": "20000331",
            "date": "2018-05-28",
            "degree": "Maths",
        },
        "proof": [
            {
                "cryptosuite": "anoncreds-2023",
                "type": "DataIntegrityProof",
                "proofPurpose": "assertionMethod",
                "verificationMethod": "LSTv7AUoTyfqFxZbuAGqKR:3:CL:536199:faber.agent.degree_schema",
                "proofValue": "ukgGEqXNjaGVtYV9pZNkvTFNUdjdBVW9UeWZxRnhaYnVBR3FLUjoyOmRlZ3JlZSBzY2hlbWE6MzUuNzguNzSrY3JlZF9kZWZfaWTZPExTVHY3QVVvVHlmcUZ4WmJ1QUdxS1I6MzpDTDo1MzYxOTk6ZmFiZXIuYWdlbnQuZGVncmVlX3NjaGVtYalzaWduYXR1cmWCrHBfY3JlZGVudGlhbISjbV8y3AAgzP_Ml8zXzOpNYcyXIxROzNHMvlk9zNnMsAkSzK7Mq0jMwxN0zMAMZMz4QszbzPkioWHcAQA3UnhCzPTM8syhS0vMjcygzJUAV8yYzLlAUcz-KszVEC0ADS7MhFMszIPMlczRf8zmd8ykzPxHzOzMkn02WA93WADM5czczOhlTk7M3szBR2jM_3bMpCnMmlIRWMy6zIQnzOkFei80ZljM_FY4zNHMn21izIBLzNDMuMyKzIYIzOh-zKF_Hcz0zNMTzNbMsczazJl4OsyBWsyMzJkrzJnMwMypzJdjzOksS05gzN7M4TfM3hrMix5HzLoyzI7Mt8z4zJIiRTHMysz-zIUidsyKzIrMn8yKzOfMuW3MoEXMsUjMzno6zNVheMyFzPMqFgk6zK1HI3HM-0bMwxcQzJDMysz-b8zMBSDMqsyQzMzMwEslF8z4zPfM5cyWzITMiMyCGQg6IzBVYszAG8z5zJtfzPVPZn5lzIzM3hctzMHM5My5F1LMhsyJzILMy1PM98z0fsyEJ1LMsMzzzN4tzMB5BVLMxsz5zNHMh8yYzKMgEzLMiUotzLLM-EYEzIshoWXcAEsQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAASzPfMkcz8fn_M4syUOszEP215zNwLoXbcAVUIzLZHzLnMyiLM0czSUjzM_EJ-zLsmzKnMncyRzO5UFTfMpcynzIE6Vsz3EVnMx1EhG8yMbszKzNxadsyKzLPM3XhcZMyWIEzM81vM1nwdE8zVzN7Mzyd4zMjM78yjG0TMuWXMiMzIF3fMyRJdK8yczLYgIi3MwczOUczRNcyuRDzMhWBGPhnMiXbMp0PMj8z8zMgiVzk_HcyDPxjMmTDM917M88z4GczIacyfzLVCR2rM21zMjloMzI44zNsPzMrMnczIfVQXzPdYzLBuzPY5zIp9cMzaBRfM0j7MoAjMucysawHM3MyrzIrMz8ymzIvMl8zzH8yCfczhzMvM5nVUTAhDUhXMkRjM717M7szWO8y8IcyYAV3M6RnM0cyrPmXMmMyKzMsEzNzM4cyTY8zTzITMqsz3HcyMLMzjfXZCK8yZXMzIMsy3zMNazJPMjMzBzMBXPHYyzI5bKsyuGczzDcy4zJvMknYTzJHMpcyYzMjMtmPM-czuzMrMycz_zLfM7MzzY8yezL3MqnnM00wXJsy9zMfM08z-zMLMugE9YszzzLsTzIQrCwPMucyqzIHMmGpozOjMrl5OzOfM7cy1KEvMqxIsP1LM-ETM_8yHzIJFUMy5zMg_zIXMwQEdGcy_zKrMh8zpzKvM0szlzNdQzPNFzLwWzMrMvQkTzK5czJU7QUYarHJfY3JlZGVudGlhbMC7c2lnbmF0dXJlX2NvcnJlY3RuZXNzX3Byb29mgqJzZdwBAFTMpszvdjHM9hBoWczmzIJiOBXM9szdzM1-zK0zcR7M9hsLzNfMuXDM4mxLzJwsYC3MuMzKacz3zK5yGRPM_QUmCMylzNfM48zvzPDMzk7MuVxtMsyZzPIXeszgzOrM_My5zL3M9czJzLR7zLrMxEAfzKsBdR_M7WPM2gvM-FYNzMtZCsytzJkra8y3aXcczODMg8z5zIIaUXrM9sySV1XM2MyCdHgazILMv3htzJ9VMjtDKxNGFMz4OsyXf8zyZHHMx8ynO09-TGp_JczPVGsSfcygG8zvGQtnbnLMzCDMycyCR23M0szca2VRW0hSzKPMrMyWzK3M8EnM_BpmQXMNGszHzIDM0MywK2sQJszjzNHM-XbMki3M68yxzON8zNR6fMy8zOgKzOZazMrMzw9nzIvMrkXM4UAGdjDMvczPIjsscFYaC8ySdRtNzOTMyDcVFsyDLMygaldrT8zwzPDMzSkHzPPM_w4wVszAoWPcACDM9Mz1zOMmzIdgfDJPQMyjzNVqF0PM3syRzLliSMyRZcy9AW3MoWTMmh9pDszs",
            }
        ],
        "issuanceDate": "2024-03-31T00:39:32.220900632Z",
    }
}


class TestV20VCDICredFormatHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        # Ledger
        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_schema = mock.CoroutineMock(return_value=SCHEMA)
        self.ledger.get_credential_definition = mock.CoroutineMock(return_value=CRED_DEF)
        self.ledger.get_revoc_reg_def = mock.CoroutineMock(return_value=REV_REG_DEF)
        self.ledger.credential_definition_id2schema_id = mock.CoroutineMock(
            return_value=SCHEMA_ID
        )
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)
        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor, mock_executor
        )
        # Context
        self.cache = InMemoryCache()
        self.profile.context.injector.bind_instance(BaseCache, self.cache)

        # Issuer
        self.issuer = mock.MagicMock(AnonCredsIssuer, autospec=True)
        self.profile.context.injector.bind_instance(AnonCredsIssuer, self.issuer)

        # Holder
        self.holder = mock.MagicMock(AnonCredsHolder, autospec=True)
        self.profile.context.injector.bind_instance(AnonCredsHolder, self.holder)

        mock_wallet = mock.MagicMock(BaseWallet, autospec=True)
        mock_wallet.get_public_did = mock.CoroutineMock(
            return_value=DIDInfo(TEST_DID, None, None, None, True)
        )
        self.profile.context.injector.bind_instance(BaseWallet, mock_wallet)

        self.handler = VCDICredFormatHandler(self.profile)

        assert self.handler.profile

    async def test_validate_fields(self):
        # Test correct data
        self.handler.validate_fields(CRED_20_PROPOSAL, {"cred_def_id": CRED_DEF_ID})
        self.handler.validate_fields(CRED_20_OFFER, VCDI_OFFER)
        # getting
        self.handler.validate_fields(CRED_20_REQUEST, VCDI_CRED_REQ)
        self.handler.validate_fields(CRED_20_ISSUE, VCDI_CRED)

        # test incorrect proposal
        with self.assertRaises(ValidationError):
            self.handler.validate_fields(
                CRED_20_PROPOSAL, {"some_random_key": "some_random_value"}
            )

        # test incorrect offer
        with self.assertRaises(ValidationError):
            offer = VCDI_OFFER.copy()
            offer.pop("binding_method")
            self.handler.validate_fields(CRED_20_OFFER, offer)

        # test incorrect request
        with self.assertRaises(ValidationError):
            req = VCDI_CRED_REQ.copy()
            req.pop("data_model_version")
            self.handler.validate_fields(CRED_20_REQUEST, req)

        # test incorrect cred
        with self.assertRaises(ValidationError):
            cred = VCDI_CRED.copy()
            cred.pop("credential")
            self.handler.validate_fields(CRED_20_ISSUE, cred)

    async def test_get_vcdi_detail_record(self):
        cred_ex_id = "dummy"
        details_vcdi = [
            V20CredExRecordIndy(
                cred_ex_id=cred_ex_id,
                rev_reg_id="rr-id",
                cred_rev_id="0",
            ),
            V20CredExRecordIndy(
                cred_ex_id=cred_ex_id,
                rev_reg_id="rr-id",
                cred_rev_id="1",
            ),
        ]
        async with self.profile.session() as session:
            await details_vcdi[0].save(session)
            await details_vcdi[1].save(session)  # exercise logger warning on get()

        with mock.patch.object(VCDI_LOGGER, "warning", mock.MagicMock()) as mock_warning:
            assert await self.handler.get_detail_record(cred_ex_id) in details_vcdi
            mock_warning.assert_called_once()

    async def test_check_uniqueness(self):
        with mock.patch.object(
            self.handler.format.detail,
            "query_by_cred_ex_id",
            mock.CoroutineMock(),
        ) as mock_vcdi_query:
            mock_vcdi_query.return_value = []
            await self.handler._check_uniqueness("dummy-cx-id")

        with mock.patch.object(
            self.handler.format.detail,
            "query_by_cred_ex_id",
            mock.CoroutineMock(),
        ) as mock_vcdi_query:
            mock_vcdi_query.return_value = [mock.MagicMock()]
            with self.assertRaises(V20CredFormatError) as context:
                await self.handler._check_uniqueness("dummy-cx-id")
            assert "detail record already exists" in str(context.exception)

    async def test_create_proposal(self):
        cred_ex_record = mock.MagicMock()
        proposal_data = {"schema_id": SCHEMA_ID}

        (cred_format, attachment) = await self.handler.create_proposal(
            cred_ex_record, proposal_data
        )

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == proposal_data

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_create_proposal_none(self):
        cred_ex_record = mock.MagicMock()
        proposal_data = None

        (cred_format, attachment) = await self.handler.create_proposal(
            cred_ex_record, proposal_data
        )

        # assert content of attachment is proposal data
        assert attachment.content == {}

    async def test_receive_proposal(self):
        cred_ex_record = mock.MagicMock()
        cred_proposal_message = mock.MagicMock()

        # Not much to assert. Receive proposal doesn't do anything
        await self.handler.receive_proposal(cred_ex_record, cred_proposal_message)

    @mock.patch.object(
        AskarWallet,
        "get_public_did",
        return_value=DIDInfo(TEST_DID, None, None, None, True),
    )
    async def test_create_offer(self, _):
        schema_id_parts = SCHEMA_ID.split(":")

        cred_preview = V20CredPreview(
            attributes=(
                V20CredAttrSpec(name="legalName", value="value"),
                V20CredAttrSpec(name="jurisdictionId", value="value"),
                V20CredAttrSpec(name="incorporationDate", value="value"),
            )
        )

        cred_proposal = V20CredProposal(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.VC_DI.api
                    ],
                )
            ],
            filters_attach=[
                AttachDecorator.data_base64({"cred_def_id": CRED_DEF_ID}, ident="0")
            ],
        )

        cred_def_record = StorageRecord(
            CRED_DEF_SENT_RECORD_TYPE,
            CRED_DEF_ID,
            {
                "schema_id": SCHEMA_ID,
                "schema_issuer_did": schema_id_parts[0],
                "schema_name": schema_id_parts[-2],
                "schema_version": schema_id_parts[-1],
                "issuer_did": TEST_DID,
                "cred_def_id": CRED_DEF_ID,
                "epoch": str(int(time())),
            },
        )
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            await storage.add_record(cred_def_record)

        with mock.patch.object(test_module, "AnonCredsIssuer", return_value=self.issuer):
            self.issuer.create_credential_offer = mock.CoroutineMock(
                return_value=json.dumps(
                    VCDI_OFFER["binding_method"]["anoncreds_link_secret"]
                )
            )

            self.issuer.match_created_credential_definitions = mock.CoroutineMock(
                return_value=CRED_DEF_ID
            )

            (cred_format, attachment) = await self.handler.create_offer(cred_proposal)

            self.issuer.create_credential_offer.assert_called_once_with(CRED_DEF_ID)

            assert cred_format.attach_id == self.handler.format.api == attachment.ident

            assert (
                attachment.content["binding_method"]["anoncreds_link_secret"]
                == VCDI_OFFER["binding_method"]["anoncreds_link_secret"]
            )

            # Assert data is encoded as base64
            assert attachment.data.base64
            # TODO: fix this get_public_did return None in the sc
            # (cred_format, attachment) = await self.handler.create_offer(cred_proposal)

            # self.issuer.create_credential_offer.assert_not_called()

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_receive_offer(self):
        cred_ex_record = mock.MagicMock()
        cred_offer_message = mock.MagicMock()

        # Not much to assert. Receive offer doesn't do anything
        await self.handler.receive_offer(cred_ex_record, cred_offer_message)

    async def test_create_request(self):
        # Define your mock credential definition
        mock_credential_definition_result = GetCredDefResult(
            credential_definition=CredDef(
                issuer_id=TEST_DID, schema_id=SCHEMA_ID, type="CL", tag="tag1", value={}
            ),
            credential_definition_id=CRED_DEF_ID,
            resolution_metadata={},
            credential_definition_metadata={},
        )
        mock_creds_registry = mock.MagicMock(AnonCredsRegistry, autospec=True)
        mock_creds_registry.get_credential_definition = mock.AsyncMock(
            return_value=mock_credential_definition_result
        )

        # Inject the MagicMock into the context
        self.profile.context.injector.bind_instance(
            AnonCredsRegistry, mock_creds_registry
        )

        holder_did = "did"

        cred_offer = V20CredOffer(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.VC_DI.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(VCDI_OFFER, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-id",
            state=V20CredExRecord.STATE_OFFER_RECEIVED,
            cred_offer=cred_offer.serialize(),
        )

        cred_req_meta = {}
        self.holder.create_credential_request = mock.CoroutineMock(
            return_value=(json.dumps(VCDI_CRED_REQ), json.dumps(cred_req_meta))
        )

        self.profile = await create_test_profile(
            {
                "wallet.type": "askar-anoncreds",
            }
        )
        with mock.patch.object(
            AnonCredsHolder, "create_credential_request", mock.CoroutineMock()
        ) as mock_create:
            mock_create.return_value = (
                json.dumps(VCDI_CRED_REQ["binding_proof"]["anoncreds_link_secret"]),
                json.dumps(cred_req_meta),
            )
            (cred_format, attachment) = await self.handler.create_request(
                cred_ex_record, {"holder_did": holder_did}
            )

            legacy_offer = await self.handler._prepare_legacy_offer(VCDI_OFFER, SCHEMA_ID)
            mock_create.assert_called_once_with(
                legacy_offer,
                mock_credential_definition_result.credential_definition,
                holder_did,
            )
            assert cred_format.attach_id == self.handler.format.api == attachment.ident

            del VCDI_CRED_REQ["binding_proof"]["anoncreds_link_secret"]["prover_did"]
            assert attachment.content == VCDI_CRED_REQ
            assert attachment.data.base64

            cred_ex_record._id = "dummy-id2"
            await self.handler.create_request(cred_ex_record, {"holder_did": holder_did})

            self.profile.context.injector.clear_binding(BaseCache)
            cred_ex_record._id = "dummy-id3"
            self.profile.context.injector.bind_instance(
                BaseMultitenantManager,
                mock.MagicMock(MultitenantManager, autospec=True),
            )
            with mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=(None, self.ledger)),
            ):
                await self.handler.create_request(
                    cred_ex_record, {"holder_did": holder_did}
                )

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_receive_request(self):
        cred_ex_record = mock.MagicMock()
        cred_request_message = mock.MagicMock()

        # Not much to assert. Receive request doesn't do anything
        await self.handler.receive_request(cred_ex_record, cred_request_message)

    async def test_issue_credential_revocable(self):
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.VC_DI.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(VCDI_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.VC_DI.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(VCDI_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        cred_rev_id = "1000"
        dummy_registry = "dummy-registry"
        expected_credential = json.dumps(VCDI_CRED)

        # Mock AnonCredsRevocation and its method create_credential_w3c
        with mock.patch.object(
            test_module, "AnonCredsRevocation", autospec=True
        ) as MockAnonCredsRevocation:
            mock_issuer = MockAnonCredsRevocation.return_value
            mock_issuer.create_credential_w3c = mock.CoroutineMock(
                return_value=(expected_credential, cred_rev_id, dummy_registry)
            )

            # Call the method under test
            (cred_format, attachment) = await self.handler.issue_credential(
                cred_ex_record, retries=1
            )
            legacy_offer = await self.handler._prepare_legacy_offer(VCDI_OFFER, SCHEMA_ID)
            legacy_request = await self.handler._prepare_legacy_request(
                VCDI_CRED_REQ, CRED_DEF_ID
            )
            # Verify the mocked method was called with the expected parameters
            mock_issuer.create_credential_w3c.assert_called_once_with(
                legacy_offer,
                legacy_request,
                attr_values,
            )

            # Assert the results are as expected
            assert cred_format.attach_id == self.handler.format.api == attachment.ident
            assert attachment.data.base64
            assert attachment.content == {"credential": VCDI_CRED}

    @pytest.mark.skip(reason="AnonCreds-break")
    async def test_issue_credential_non_revocable(self):
        CRED_DEF_NR = deepcopy(CRED_DEF)
        CRED_DEF_NR["value"]["revocation"] = None
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.VC_DI.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(VCDI_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.VC_DI.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(VCDI_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecordIndy(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecordIndy.INITIATOR_SELF,
            role=V20CredExRecordIndy.ROLE_ISSUER,
            state=V20CredExRecordIndy.STATE_REQUEST_RECEIVED,
        )

        self.issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(VCDI_CRED), None)
        )
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value=CRED_DEF_NR
        )
        self.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor,
            "get_ledger_for_identifier",
            mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
        ):
            (cred_format, attachment) = await self.handler.issue_credential(
                cred_ex_record, retries=0
            )

            self.issuer.create_credential.assert_called_once_with(
                SCHEMA,
                VCDI_OFFER,
                VCDI_CRED_REQ,
                attr_values,
                None,
                None,
            )

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == VCDI_CRED

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_match_sent_cred_def_id_error(self):
        tag_query = {"tag": "test_tag"}

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler._match_sent_cred_def_id(tag_query)
        assert "Issuer has no operable cred def for proposal spec " in str(
            context.exception
        )

    async def test_store_credential(self):
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.VC_DI.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(VCDI_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.VC_DI.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(VCDI_CRED_REQ, ident="0")],
        )
        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.VC_DI.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(VCDI_CRED, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            cred_issue=cred_issue.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )
        cred_id = "dummy-cred-id"

        # Define your mock credential definition
        mock_credential_definition_result = GetCredDefResult(
            credential_definition=CredDef(
                issuer_id=TEST_DID, schema_id=SCHEMA_ID, type="CL", tag="tag1", value={}
            ),
            credential_definition_id=CRED_DEF_ID,
            resolution_metadata={},
            credential_definition_metadata={},
        )
        mock_creds_registry = mock.MagicMock(AnonCredsRegistry, autospec=True)
        mock_creds_registry.get_credential_definition = mock.AsyncMock(
            return_value=mock_credential_definition_result
        )

        revocation_registry = RevRegDef(
            cred_def_id=CRED_DEF_ID,
            issuer_id=TEST_DID,
            tag="tag1",
            type="CL_ACCUM",
            value={},
        )

        mock_creds_registry.get_revocation_registry_definition = mock.AsyncMock(
            return_value=GetRevRegDefResult(
                revocation_registry=revocation_registry,
                revocation_registry_id="rr-id",
                resolution_metadata={},
                revocation_registry_metadata={},
            )
        )
        # Inject the MagicMock into the context
        self.profile.context.injector.bind_instance(
            AnonCredsRegistry, mock_creds_registry
        )
        self.profile = await create_test_profile(
            {
                "wallet.type": "askar-anoncreds",
            }
        )
        with mock.patch.object(
            test_module.AnonCredsRevocation,
            "get_or_fetch_local_tails_path",
            mock.CoroutineMock(),
        ):
            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.store_credential(cred_ex_record, cred_id)
            assert (
                "No credential exchange didcomm/ detail record found for cred ex id dummy-cxid"
                in str(context.exception)
            )

            record = V20CredExRecordIndy(
                cred_ex_indy_id="dummy-cxid",
                rev_reg_id="rr-id",
                cred_ex_id="dummy-cxid",
                cred_id_stored=cred_id,
                cred_request_metadata="dummy-metadata",
                cred_rev_id="0",
            )

            record.save = mock.CoroutineMock()
            self.handler.get_detail_record = mock.AsyncMock(return_value=record)
            with mock.patch.object(
                W3cCredential,
                "rev_reg_id",
                return_value="rr-id",
            ):
                with mock.patch.object(
                    AnonCredsHolder,
                    "store_credential_w3c",
                    mock.AsyncMock(),
                ) as mock_store_credential:
                    # Error case: no cred ex record found

                    await self.handler.store_credential(cred_ex_record, cred_id)

                    mock_store_credential.assert_called_once_with(
                        mock_credential_definition_result.credential_definition.serialize(),
                        VCDI_CRED["credential"],
                        record.cred_request_metadata,
                        None,
                        credential_id=cred_id,
                        rev_reg_def=revocation_registry.serialize(),
                    )

            with mock.patch.object(
                AnonCredsHolder,
                "store_credential_w3c",
                mock.AsyncMock(side_effect=AnonCredsHolderError),
            ) as mock_store_credential:
                with self.assertRaises(AnonCredsHolderError) as context:
                    await self.handler.store_credential(cred_ex_record, cred_id)



from unittest import IsolatedAsyncioTestCase

from ....tests import mock
from ....utils.testing import create_test_profile
from ...base import BaseLedger
from ...indy_vdr import IndyVdrLedger, IndyVdrLedgerPool
from ...multiple_ledger.base_manager import (
    BaseMultipleLedgerManager,
    MultipleLedgerManagerError,
)
from ..ledger_requests_executor import IndyLedgerRequestsExecutor


class TestIndyLedgerRequestsExecutor(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.context = self.profile.context
        setattr(self.context, "profile", self.profile)
        self.profile.settings["ledger.ledger_config_list"] = [
            {
                "id": "test_prod_1",
                "pool_name": "test_prod_1",
                "is_production": True,
                "genesis_transactions": "genesis_transactions",
            }
        ]
        self.ledger = IndyVdrLedger(IndyVdrLedgerPool("test_prod_1"), self.profile)
        mock_ledger_manger = mock.MagicMock(BaseMultipleLedgerManager, autospec=True)
        mock_ledger_manger.extract_did_from_identifier = mock.MagicMock(
            return_value="WgWxqztrNooG92RXvxSTWv"
        )
        mock_ledger_manger.lookup_did_in_configured_ledgers = mock.CoroutineMock(
            return_value=("test_prod_1", self.ledger)
        )
        mock_ledger_manger.get_ledger_inst_by_id = mock.CoroutineMock(
            return_value=self.ledger
        )
        self.profile.context.injector.bind_instance(
            BaseMultipleLedgerManager, mock_ledger_manger
        )
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)
        self.indy_ledger_requestor = IndyLedgerRequestsExecutor(self.profile)

    async def test_get_ledger_for_identifier(self):
        (
            ledger_id,
            ledger_inst,
        ) = await self.indy_ledger_requestor.get_ledger_for_identifier(
            "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0", 0
        )
        assert ledger_id == "test_prod_1"
        assert ledger_inst.pool.name == "test_prod_1"

    async def test_get_ledger_inst(self):
        ledger_inst = await self.indy_ledger_requestor.get_ledger_inst("test_prod_1")
        assert ledger_inst

    async def test_get_ledger_for_identifier_is_digit(self):
        ledger_id, ledger = await self.indy_ledger_requestor.get_ledger_for_identifier(
            "123", 0
        )
        assert ledger_id is None
        assert ledger == self.ledger

    async def test_get_ledger_for_identifier_x(self):
        self.profile.context.injector.bind_instance(
            BaseMultipleLedgerManager,
            mock.MagicMock(
                extract_did_from_identifier=mock.MagicMock(
                    return_value="WgWxqztrNooG92RXvxSTWv"
                ),
                lookup_did_in_configured_ledgers=mock.CoroutineMock(
                    side_effect=MultipleLedgerManagerError
                ),
            ),
        )
        self.indy_ledger_requestor = IndyLedgerRequestsExecutor(self.profile)
        ledger_id, ledger = await self.indy_ledger_requestor.get_ledger_for_identifier(
            "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0", 0
        )
        assert ledger_id is None
        assert ledger == self.ledger

    async def test_get_ledger_for_identifier_mult_ledger_not_set(self):
        self.profile.settings["ledger.ledger_config_list"] = None
        self.indy_ledger_requestor = IndyLedgerRequestsExecutor(self.profile)
        ledger_id, ledger = await self.indy_ledger_requestor.get_ledger_for_identifier(
            "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0", 0
        )
        assert ledger_id is None
        assert ledger == self.ledger

    async def test_get_ledger_for_identifier_mult_ledger_not_cached(self):
        (
            ledger_id,
            ledger_inst,
        ) = await self.indy_ledger_requestor.get_ledger_for_identifier(
            "GUTK6XARozQCWxqzPSUr4g", 4
        )
        assert ledger_id == "test_prod_1"
        assert ledger_inst.pool.name == "test_prod_1"



"""Test Problem Report Handler."""

import pytest
import pytest_asyncio

from ......connections.models.conn_record import ConnRecord
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...handlers import problem_report_handler as test_module
from ...messages.problem_report import CMProblemReport, ProblemReportReason


@pytest_asyncio.fixture
async def request_context():
    ctx = RequestContext.test_context(await create_test_profile())
    ctx.message_receipt = MessageReceipt()
    yield ctx


@pytest_asyncio.fixture
async def connection_record(request_context, session):
    record = ConnRecord()
    request_context.connection_record = record
    await record.save(session)
    yield record


@pytest_asyncio.fixture
async def session(request_context):
    yield await request_context.session()


class TestCMProblemReportHandler:
    @pytest.mark.asyncio
    async def test_cover(self, request_context, connection_record):
        request_context.message = CMProblemReport(
            description={
                "en": "Mediation not granted",
                "code": ProblemReportReason.MEDIATION_NOT_GRANTED.value,
            }
        )
        handler = test_module.CMProblemReportHandler()
        responder = MockResponder()
        await handler.handle(context=request_context, responder=responder)



from unittest import TestCase

from ......messaging.decorators.attach_decorator import AttachDecorator
from ..pres_format import V20PresFormat

CD_ID = "GMm4vMw8LLrLJjp81kRRLp:3:CL:12:tag"
INDY_PROOF_REQ = [
    {
        "name": "proof-req",
        "version": "1.0",
        "nonce": "12345",
        "requested_attributes": {
            "0_player_uuid": {
                "name": "player",
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
                "non_revoked": {
                    "from": 1234567890,
                    "to": 1234567890,
                },
            },
            "0_screencapture_uuid": {
                "name": "screenCapture",
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
                "non_revoked": {
                    "from": 1234567890,
                    "to": 1234567890,
                },
            },
        },
        "requested_predicates": {
            "0_highscore_GE_uuid": {
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
                "non_revoked": {
                    "from": 1234567890,
                    "to": 1234567890,
                },
            }
        },
    },
    {
        "name": "proof-req",
        "version": "1.0",
        "nonce": "123456",
        "requested_attributes": {
            "0_player_uuid": {
                "name": "player",
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            },
            "0_screencapture_uuid": {
                "name": "screenCapture",
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            },
        },
        "requested_predicates": {
            "0_highscore_GE_uuid": {
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            }
        },
    },
    {
        "name": "proof-req",
        "version": "1.0",
        "nonce": "1234567",
        "requested_attributes": {},
        "requested_predicates": {
            "0_highscore_GE_uuid": {
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            }
        },
    },
]


class TestV20FormatFormat(TestCase):
    """Coverage for self-get."""

    def test_get_completeness(self):
        assert (
            V20PresFormat.Format.get(V20PresFormat.Format.INDY)
            is V20PresFormat.Format.INDY
        )
        assert V20PresFormat.Format.get("no such format") is None
        assert V20PresFormat.Format.get("hlindy/...") is V20PresFormat.Format.INDY
        assert (
            V20PresFormat.Format.get(V20PresFormat.Format.DIF.api)
            is V20PresFormat.Format.DIF
        )

    def test_get_attachment_data(self):
        for proof_req in INDY_PROOF_REQ:
            assert (
                V20PresFormat.Format.INDY.get_attachment_data(
                    formats=[
                        V20PresFormat(
                            attach_id="indy",
                            format_=V20PresFormat.Format.INDY,
                        )
                    ],
                    attachments=[
                        AttachDecorator.data_base64(mapping=proof_req, ident="indy")
                    ],
                )
                == proof_req
            )

            assert (
                V20PresFormat.Format.INDY.get_attachment_data(
                    formats=[
                        V20PresFormat(
                            attach_id="indy",
                            format_=V20PresFormat.Format.INDY,
                        )
                    ],
                    attachments=[AttachDecorator.data_base64(proof_req, ident="xxx")],
                )
                is None
            )

            assert (
                V20PresFormat.Format.DIF.get_attachment_data(
                    formats=[
                        V20PresFormat(
                            attach_id="indy",
                            format_=V20PresFormat.Format.INDY,
                        )
                    ],
                    attachments=[AttachDecorator.data_base64(proof_req, ident="indy")],
                )
                is None
            )



from unittest import IsolatedAsyncioTestCase

import pytest

from acapy_agent.resolver.did_resolver import DIDResolver
from acapy_agent.wallet.key_type import KeyTypes
from acapy_agent.wallet.keys.manager import MultikeyManager

from ...did.did_key import DIDKey
from ...resolver.tests.test_did_resolver import MockResolver
from ...utils.testing import create_test_profile
from ...wallet.default_verification_key_strategy import DefaultVerificationKeyStrategy

TEST_DID_SOV = "did:sov:LjgpST2rjsoxYegQDRm7EL"
TEST_SEED = "testseed000000000000000000000001"
TEST_ED25519_MULTIKEY = "z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
TEST_ED25519_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
TEST_BLS_G2_MULTIKEY = "zUC74E9UD2W6Q1MgPexCEdpstiCsY1Vbnyqepygk7McZVce38L1tGX7qZ2SgY4Zz2m9FUB4Xb5cEHSujks9XeKDzqe4QzW3CyyJ1cv8iBLNqU61EfkBoW2yEkg6VgqHTDtANYRS"
TEST_BLS_G2_VERKEY = "pPbb9Lqs3PVTyiHM4h8fbQqxHjBPm1Hixb6vdW9kkjHEij4FZrigkaV1P5DjWTbcKxeeYfkQuZMmozRQV3tH1gXhCA972LAXMGSKH7jxz8sNJqrCR6o8asgXDeYZeL1W3p8"
TEST_DID_KEY = f"did:key:{TEST_ED25519_MULTIKEY}"


class TestDefaultVerificationKeyStrategy(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile()
        resolver = DIDResolver()
        resolver.register_resolver(
            MockResolver(
                ["example"],
                resolved={
                    "@context": [
                        "https://www.w3.org/ns/did/v1",
                        "https://w3id.org/security/multikey/v1",
                    ],
                    "id": "did:example:123",
                    "verificationMethod": [
                        # VM has a key not owned by this acapy agent
                        {
                            "id": "did:example:123#not-owned",
                            "type": "Multikey",
                            "controller": "did:example:123",
                            "publicKeyMultibase": "z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDHHyGo38EefXmgDL",
                        },
                        {
                            "id": "did:example:123#key-1",
                            "type": "Multikey",
                            "controller": "did:example:123",
                            "publicKeyMultibase": TEST_ED25519_MULTIKEY,
                        },
                        {
                            "id": "did:example:123#key-2",
                            "type": "Multikey",
                            "controller": "did:example:123",
                            "publicKeyMultibase": TEST_ED25519_MULTIKEY,
                        },
                        {
                            "id": "did:example:123#key-3",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:example:123",
                            "publicKeyBase58": TEST_ED25519_VERKEY,
                        },
                    ],
                    "authentication": ["did:example:123#key-1"],
                    "assertionMethod": [
                        "did:example:123#not-owned",
                        "did:example:123#key-2",
                        "did:example:123#key-3",
                        {
                            "id": "did:example:123#key-4",
                            "type": "Bls12381G2Key2020",
                            "controller": "did:example:123",
                            "publicKeyBase58": TEST_BLS_G2_VERKEY,
                        },
                    ],
                },
            )
        )
        self.profile.context.injector.bind_instance(DIDResolver, resolver)
        self.profile.context.injector.bind_instance(KeyTypes, KeyTypes())
        async with self.profile.session() as session:
            await MultikeyManager(session=session).create(seed=TEST_SEED, alg="ed25519")
            await MultikeyManager(session=session).create(
                seed=TEST_SEED, alg="bls12381g2"
            )

    async def test_with_did_sov(self):
        strategy = DefaultVerificationKeyStrategy()
        assert (
            await strategy.get_verification_method_id_for_did(TEST_DID_SOV, self.profile)
            == TEST_DID_SOV + "#key-1"
        )
        with pytest.raises(Exception):
            await strategy.get_verification_method_id_for_did(
                did=TEST_DID_SOV,
                profile=self.profile,
                proof_type="BbsBlsSignature2020",
            )
        with pytest.raises(Exception):
            await strategy.get_verification_method_id_for_did(
                did=TEST_DID_SOV,
                profile=self.profile,
                verification_method_id=f"{TEST_DID_KEY}#key-2",
            )

    async def test_with_did_key(self):
        strategy = DefaultVerificationKeyStrategy()
        assert (
            await strategy.get_verification_method_id_for_did(TEST_DID_KEY, self.profile)
            == DIDKey.from_did(TEST_DID_KEY).key_id
        )
        with pytest.raises(Exception):
            await strategy.get_verification_method_id_for_did(
                did=TEST_DID_KEY,
                profile=self.profile,
                proof_type="BbsBlsSignature2020",
            )
        with pytest.raises(Exception):
            await strategy.get_verification_method_id_for_did(
                did=TEST_DID_KEY,
                profile=self.profile,
                verification_method_id=f"{TEST_DID_KEY}#abc",
            )

    async def test_with_did_for_assertion(self):
        strategy = DefaultVerificationKeyStrategy()
        assert (
            await strategy.get_verification_method_id_for_did(
                "did:example:123",
                self.profile,
                proof_type="Ed25519Signature2020",
                proof_purpose="assertionMethod",
            )
            == "did:example:123#key-2"
        )
        assert (
            await strategy.get_verification_method_id_for_did(
                "did:example:123",
                self.profile,
                proof_type="Ed25519Signature2018",
                proof_purpose="assertionMethod",
            )
            == "did:example:123#key-2"
        )
        assert (
            await strategy.get_verification_method_id_for_did(
                "did:example:123",
                self.profile,
                proof_type="Ed25519Signature2018",
                proof_purpose="assertionMethod",
                verification_method_id="did:example:123#key-3",
            )
            == "did:example:123#key-3"
        )
        assert (
            await strategy.get_verification_method_id_for_did(
                "did:example:123",
                self.profile,
                proof_type="BbsBlsSignature2020",
                proof_purpose="assertionMethod",
            )
            == "did:example:123#key-4"
        )

    async def test_fail_cases(self):
        strategy = DefaultVerificationKeyStrategy()
        # base case
        assert (
            await strategy.get_verification_method_id_for_did(
                "did:example:123",
                self.profile,
                proof_type="Ed25519Signature2020",
                proof_purpose="assertionMethod",
                verification_method_id="did:example:123#key-2",
            )
            == "did:example:123#key-2"
        )
        with pytest.raises(Exception):
            # nothing suitable for purpose
            await strategy.get_verification_method_id_for_did(
                "did:example:123",
                self.profile,
                proof_type="Ed25519Signature2020",
                proof_purpose="capabilityInvocation",
            )
        with pytest.raises(Exception):
            # nothing suitable for proof type
            await strategy.get_verification_method_id_for_did(
                "did:example:123",
                self.profile,
                proof_type="EcdsaSecp256r1Signature2019",
                proof_purpose="assertionMethod",
            )
        with pytest.raises(Exception):
            # suitable, but key not owned by acapy
            await strategy.get_verification_method_id_for_did(
                "did:example:123",
                self.profile,
                proof_type="Ed25519Signature2020",
                proof_purpose="assertionMethod",
                verification_method_id="did:example:123#not-owned",
            )

    async def test_unsupported_did_method(self):
        strategy = DefaultVerificationKeyStrategy()
        with pytest.raises(Exception):
            await strategy.get_verification_method_id_for_did(
                "did:test:test", self.profile
            )



"""Test mediation record."""

import json

import pytest
import pytest_asyncio

from ......core.profile import ProfileSession
from ......storage.base import BaseStorage
from ......storage.record import StorageRecord
from ......utils.testing import create_test_profile
from ..mediation_record import MediationRecord


@pytest_asyncio.fixture
async def session():
    profile = await create_test_profile()
    async with profile.session() as session:
        yield session


@pytest.mark.asyncio
async def test_backwards_compat_terms(session: ProfileSession):
    """Make sure old records can be loaded still."""

    old_record = StorageRecord(
        MediationRecord.RECORD_TYPE,
        id="test_mediation_id",
        value=json.dumps(
            {
                "state": "granted",
                "mediator_terms": ["dummy_terms"],
                "recipient_terms": ["dummy_terms"],
            }
        ),
    )
    storage = session.inject(BaseStorage)
    await storage.add_record(old_record)

    record = await MediationRecord.retrieve_by_id(session, old_record.id)
    assert isinstance(record, MediationRecord)
    assert record.mediation_id == "test_mediation_id"
    assert record.state == "granted"


@pytest.mark.asyncio
async def test_mediation_record_eq():
    record_0 = MediationRecord(mediation_id="test_mediation_id_0", endpoint="zero")
    record_1 = MediationRecord(mediation_id="test_mediation_id_1", endpoint="one")
    assert record_0 != record_1

    with pytest.raises(ValueError):
        record_0.state = "bad state"


@pytest.mark.asyncio
async def test_mediation_record_duplicate_means_exists(session: ProfileSession):
    await MediationRecord(connection_id="test_connection_id", endpoint="abc").save(
        session
    )
    await MediationRecord(connection_id="test_connection_id", endpoint="def").save(
        session
    )
    assert await MediationRecord.exists_for_connection_id(session, "test_connection_id")



"""Schema for configuring multiple ledgers."""

from typing import Optional

from marshmallow import EXCLUDE, fields, pre_load
from uuid_utils import uuid4

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.models.openapi import OpenAPISchema


class LedgerConfigInstance(BaseModel):
    """describes each LedgerConfigInstance for multiple ledger support."""

    class Meta:
        """LedgerConfigInstance metadata."""

        schema_class = "LedgerConfigInstanceSchema"

    def __init__(
        self,
        *,
        id: Optional[str] = None,
        is_production: bool = True,
        is_write: bool = False,
        keepalive: int = 5,
        read_only: bool = False,
        socks_proxy: Optional[str] = None,
        pool_name: Optional[str] = None,
        endorser_alias: Optional[str] = None,
        endorser_did: Optional[str] = None,
    ):
        """Initialize LedgerConfigInstance."""
        self.id = id or str(uuid4())
        self.is_production = is_production
        self.is_write = is_write
        self.keepalive = keepalive
        self.read_only = read_only
        self.socks_proxy = socks_proxy
        self.pool_name = pool_name or self.id
        self.endorser_alias = endorser_alias
        self.endorser_did = endorser_did


class LedgerConfigInstanceSchema(BaseModelSchema):
    """Single LedgerConfigInstance Schema."""

    class Meta:
        """LedgerConfigInstanceSchema metadata."""

        model_class = LedgerConfigInstance
        unknown = EXCLUDE

    id = fields.Str(
        required=True,
        metadata={
            "description": "Ledger identifier. Auto-generated UUID4 if not provided",
            "example": "f47ac10b-58cc-4372-a567-0e02b2c3d479",
        },
    )
    is_production = fields.Bool(
        required=True, metadata={"description": "Production-grade ledger (true/false)"}
    )
    is_write = fields.Bool(
        required=False,
        metadata={"description": "Write capability enabled (default: False)"},
    )
    keepalive = fields.Int(
        required=False,
        metadata={
            "description": "Keep-alive timeout in seconds for idle connections",
            "default": 5,
        },
    )
    read_only = fields.Bool(
        required=False, metadata={"description": "Read-only access (default: False)"}
    )
    socks_proxy = fields.Str(
        required=False, metadata={"description": "SOCKS proxy URL (optional)"}
    )
    pool_name = fields.Str(
        required=False,
        metadata={
            "description": "Ledger pool name (defaults to ledger ID if not specified)",
            "example": "bcovrin-test-pool",
        },
    )
    endorser_alias = fields.Str(
        required=False, metadata={"description": "Endorser service alias (optional)"}
    )
    endorser_did = fields.Str(
        required=False, metadata={"description": "Endorser DID (optional)"}
    )

    @pre_load
    def validate_id(self, data, **kwargs):
        """Check if id is present, if not then set to UUID4."""
        if "id" not in data:
            data["id"] = str(uuid4())
        return data

    @pre_load
    def set_defaults(self, data, **kwargs):
        """Set default values for optional fields."""
        data.setdefault("is_write", False)
        data.setdefault("keepalive", 5)
        data.setdefault("read_only", False)
        return data


class LedgerConfigListSchema(OpenAPISchema):
    """Schema for Ledger Config List."""

    production_ledgers = fields.List(  # Changed from ledger_config_list
        fields.Nested(LedgerConfigInstanceSchema(), required=True),
        required=True,
        metadata={"description": "Production ledgers (may be empty)"},
    )
    non_production_ledgers = fields.List(  # Added new field
        fields.Nested(LedgerConfigInstanceSchema(), required=True),
        required=True,
        metadata={"description": "Non-production ledgers (may be empty)"},
    )


class WriteLedgerSchema(OpenAPISchema):
    """Schema for getting ledger_id for the write ledger."""

    ledger_id = fields.Str()


class ConfigurableWriteLedgersSchema(OpenAPISchema):
    """Schema for list of configurable write ledger."""

    write_ledgers = fields.List(
        fields.Str(metadata={"description": "Ledgers identifiers"}),
        metadata={"description": "List of configurable write ledgers identifiers"},
    )


class MultipleLedgerModuleResultSchema(OpenAPISchema):
    """Schema for the multiple ledger modules endpoint."""



"""Classes representing inbound messages."""

import asyncio
from typing import Optional, Union

from .receipt import MessageReceipt


class InboundMessage:
    """Container class linking a message payload with its receipt details."""

    def __init__(
        self,
        payload: Union[str, bytes],
        receipt: MessageReceipt,
        *,
        connection_id: Optional[str] = None,
        session_id: Optional[str] = None,
        transport_type: Optional[str] = None,
    ):
        """Initialize the inbound message."""
        self.connection_id = connection_id
        self.payload = payload
        self.receipt = receipt
        self.session_id = session_id
        self.transport_type = transport_type
        self.processing_complete_event = asyncio.Event()

    def dispatch_processing_complete(self):
        """Dispatch processing complete."""
        self.processing_complete_event.set()

    async def wait_processing_complete(self):
        """Wait for processing to complete."""
        await self.processing_complete_event.wait()



from unittest import IsolatedAsyncioTestCase

from ....connections.models.connection_target import ConnectionTarget
from ....transport.outbound.message import OutboundMessage
from ..delivery_queue import DeliveryQueue


class TestDeliveryQueue(IsolatedAsyncioTestCase):
    async def test_message_add_and_check(self):
        queue = DeliveryQueue()

        t = ConnectionTarget(recipient_keys=["aaa"])
        msg = OutboundMessage(payload="x", target=t)
        queue.add_message(msg)
        assert queue.has_message_for_key("aaa")

    async def test_message_add_not_false_check(self):
        queue = DeliveryQueue()

        t = ConnectionTarget(recipient_keys=["aaa"])
        msg = OutboundMessage(payload="x", target=t)
        queue.add_message(msg)
        assert queue.has_message_for_key("bbb") is False

    async def test_message_add_get_by_key(self):
        queue = DeliveryQueue()

        t = ConnectionTarget(recipient_keys=["aaa"])
        msg = OutboundMessage(payload="x", target=t)
        queue.add_message(msg)
        assert queue.has_message_for_key("aaa")
        assert queue.get_one_message_for_key("aaa") == msg
        assert queue.has_message_for_key("aaa") is False

    async def test_message_add_get_by_list(self):
        queue = DeliveryQueue()

        t = ConnectionTarget(recipient_keys=["aaa"])
        msg = OutboundMessage(payload="x", target=t)
        queue.add_message(msg)
        assert queue.has_message_for_key("aaa")
        msg_list = list(queue.inspect_all_messages_for_key("aaa"))
        assert queue.message_count_for_key("aaa") == 1
        assert len(msg_list) == 1
        assert msg_list[0] == msg
        queue.remove_message_for_key("aaa", msg)
        assert queue.has_message_for_key("aaa") is False

    async def test_message_ttl(self):
        queue = DeliveryQueue()

        t = ConnectionTarget(recipient_keys=["aaa"])
        msg = OutboundMessage(payload="x", target=t)
        queue.add_message(msg)
        assert queue.has_message_for_key("aaa")
        queue.expire_messages(ttl=-10)
        assert queue.has_message_for_key("aaa") is False

    async def test_count_zero_with_no_items(self):
        queue = DeliveryQueue()
        assert queue.message_count_for_key("aaa") == 0



"""DIF Proof Request Schema."""

from typing import Optional, Union

from marshmallow import INCLUDE, fields

from ....messaging.models.base import BaseModel, BaseModelSchema
from ....messaging.models.openapi import OpenAPISchema
from .pres_exch import (
    DIFOptions,
    DIFOptionsSchema,
    PresentationDefinition,
    PresentationDefinitionSchema,
)


class DIFProofRequest(BaseModel):
    """DIF presentation request input detail."""

    class Meta:
        """DIFProofRequest metadata."""

        schema_class = "DIFProofRequestSchema"

    def __init__(
        self,
        presentation_definition: Optional[Union[dict, PresentationDefinition]],
        options: Optional[Union[dict, DIFOptions]] = None,
    ) -> None:
        """Initialize the DIFProofRequest instance."""
        self.presentation_definition = presentation_definition
        self.options = options


class DIFProofRequestSchema(BaseModelSchema):
    """Schema for DIF presentation request."""

    class Meta:
        """Accept parameter overload."""

        unknown = INCLUDE
        model_class = DIFProofRequest

    options = fields.Nested(DIFOptionsSchema(), required=False)
    presentation_definition = fields.Nested(PresentationDefinitionSchema(), required=True)


class DIFPresSpecSchema(OpenAPISchema):
    """Schema for DIF Presentation Spec schema."""

    issuer_id = fields.Str(
        required=False,
        metadata={
            "description": (
                "Issuer identifier to sign the presentation, if different from current"
                " public DID"
            )
        },
    )
    record_ids = fields.Dict(
        required=False,
        metadata={
            "description": (
                "Mapping of input_descriptor id to list of stored W3C credential"
                " record_id"
            ),
            "example": {
                "<input descriptor id_1>": ["<record id_1>", "<record id_2>"],
                "<input descriptor id_2>": ["<record id>"],
            },
        },
    )
    presentation_definition = fields.Nested(
        PresentationDefinitionSchema(), required=False
    )
    reveal_doc = fields.Dict(
        required=False,
        metadata={
            "description": (
                "reveal doc [JSON-LD frame] dict used to derive the credential when"
                " selective disclosure is required"
            ),
            "example": {
                "@context": [
                    "https://www.w3.org/2018/credentials/v1",
                    "https://w3id.org/security/bbs/v1",
                ],
                "type": ["VerifiableCredential", "LabReport"],
                "@explicit": True,
                "@requireAll": True,
                "issuanceDate": {},
                "issuer": {},
                "credentialSubject": {
                    "Observation": [
                        {
                            "effectiveDateTime": {},
                            "@explicit": True,
                            "@requireAll": True,
                        }
                    ],
                    "@explicit": True,
                    "@requireAll": True,
                },
            },
        },
    )



from unittest import IsolatedAsyncioTestCase, TestCase, mock

from ......did.did_key import DIDKey
from ......wallet.key_type import ED25519
from .....didcomm_prefix import DIDCommPrefix
from .....out_of_band.v1_0.messages.invitation import InvitationMessage, Service
from ...message_types import FORWARD_INVITATION, PROTOCOL_PACKAGE
from ..forward_invitation import ForwardInvitation


class TestConfig:
    label = "Label"
    did = "did:sov:QmWbsNYhMrjHiqZDTUTEJs"
    endpoint_url = "https://example.com/endpoint"
    endpoint_did = "did:sov:A2wBhNYhMrjHiqZDTUYH7u"
    key = "8HH5gYEeNc3z7PYXmd54d4x6qAfCNrqQqEB3nS7Zfu7K"
    did_key = DIDKey.from_public_key_b58(key, ED25519)
    test_message = "test message"


class TestForwardInvitation(TestCase, TestConfig):
    def setUp(self):
        self.service = Service(
            recipient_keys=[self.key], service_endpoint=self.endpoint_url
        )
        self.connection_invitation = InvitationMessage(
            label=self.label, services=[self.service]
        )
        self.invitation = ForwardInvitation(
            invitation=self.connection_invitation, message=self.test_message
        )

    def test_init(self):
        """Test initialization."""
        assert self.invitation.invitation == self.connection_invitation
        assert self.invitation.message == self.test_message

    def test_type(self):
        """Test type."""
        assert self.invitation._type == DIDCommPrefix.qualify_current(FORWARD_INVITATION)

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.forward_invitation.ForwardInvitationSchema.load"
    )
    def test_deserialize(self, mock_invitation_schema_load):
        """
        Test deserialization.
        """
        obj = {"obj": "obj"}

        invitation = ForwardInvitation.deserialize(obj)
        mock_invitation_schema_load.assert_called_once_with(obj)

        assert invitation is mock_invitation_schema_load.return_value

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.forward_invitation.ForwardInvitationSchema.dump"
    )
    def test_serialize(self, mock_invitation_schema_dump):
        """
        Test serialization.
        """
        invitation_dict = self.invitation.serialize()
        mock_invitation_schema_dump.assert_called_once_with(self.invitation)

        assert invitation_dict is mock_invitation_schema_dump.return_value


class TestForwardInvitationSchema(IsolatedAsyncioTestCase, TestConfig):
    """Test forward invitation schema."""

    async def test_make_model(self):
        service = Service(
            _id="asdf",
            _type="did-communication",
            recipient_keys=[self.did_key.did],
            service_endpoint=self.endpoint_url,
        )
        invitation = ForwardInvitation(
            invitation=InvitationMessage(
                label=self.label,
                services=[service],
                handshake_protocols=["didexchange/1.1"],
            ),
            message=self.test_message,
        )
        data = invitation.serialize()
        model_instance = ForwardInvitation.deserialize(data)
        assert type(model_instance) is type(invitation)



"""DID Web Registry."""

import logging
import re
from typing import Optional, Pattern, Sequence

from ....config.injection_context import InjectionContext
from ....core.profile import Profile
from ...base import BaseAnonCredsRegistrar, BaseAnonCredsResolver
from ...models.credential_definition import CredDef, CredDefResult, GetCredDefResult
from ...models.revocation import (
    GetRevListResult,
    GetRevRegDefResult,
    RevList,
    RevListResult,
    RevRegDef,
    RevRegDefResult,
)
from ...models.schema import AnonCredsSchema, GetSchemaResult, SchemaResult
from ...models.schema_info import AnonCredsSchemaInfo

LOGGER = logging.getLogger(__name__)


class DIDWebRegistry(BaseAnonCredsResolver, BaseAnonCredsRegistrar):
    """DIDWebRegistry."""

    def __init__(self):
        """Initialize an instance.

        Args:
            None

        """
        self._supported_identifiers_regex = re.compile(
            r"^did:web:[a-z0-9]+(?:\.[a-z0-9]+)*(?::\d+)?(?:\/[^#\s]*)?(?:#.*)?\s*$"
        )

    @property
    def supported_identifiers_regex(self) -> Pattern:
        """Supported Identifiers Regular Expression."""
        return self._supported_identifiers_regex
        # TODO: fix regex (too general)

    async def setup(self, context: InjectionContext) -> None:
        """Setup."""
        LOGGER.info("Successfully registered DIDWebRegistry")

    async def get_schema(self, profile: Profile, schema_id: str) -> GetSchemaResult:
        """Get a schema from the registry."""
        raise NotImplementedError()

    async def register_schema(
        self,
        profile: Profile,
        schema: AnonCredsSchema,
        options: Optional[dict] = None,
    ) -> SchemaResult:
        """Register a schema on the registry."""
        raise NotImplementedError()

    async def get_credential_definition(
        self, profile: Profile, credential_definition_id: str
    ) -> GetCredDefResult:
        """Get a credential definition from the registry."""
        raise NotImplementedError()

    async def register_credential_definition(
        self,
        profile: Profile,
        schema: GetSchemaResult,
        credential_definition: CredDef,
        options: Optional[dict] = None,
    ) -> CredDefResult:
        """Register a credential definition on the registry."""
        raise NotImplementedError()

    async def get_revocation_registry_definition(
        self, profile: Profile, revocation_registry_id: str
    ) -> GetRevRegDefResult:
        """Get a revocation registry definition from the registry."""
        raise NotImplementedError()

    async def register_revocation_registry_definition(
        self,
        profile: Profile,
        revocation_registry_definition: RevRegDef,
        options: Optional[dict] = None,
    ) -> RevRegDefResult:
        """Register a revocation registry definition on the registry."""
        raise NotImplementedError()

    async def get_revocation_list(
        self,
        profile: Profile,
        revocation_registry_id: str,
        timestamp_from: Optional[int] = 0,
        timestamp_to: Optional[int] = None,
    ) -> GetRevListResult:
        """Get a revocation list from the registry."""
        raise NotImplementedError()

    async def register_revocation_list(
        self,
        profile: Profile,
        rev_reg_def: RevRegDef,
        rev_list: RevList,
        options: Optional[dict] = None,
    ) -> RevListResult:
        """Register a revocation list on the registry."""
        raise NotImplementedError()

    async def update_revocation_list(
        self,
        profile: Profile,
        rev_reg_def: RevRegDef,
        prev_list: RevList,
        curr_list: RevList,
        revoked: Sequence[int],
        options: Optional[dict] = None,
    ) -> RevListResult:
        """Update a revocation list on the registry."""
        raise NotImplementedError()

    async def get_schema_info_by_id(
        self, profile: Profile, schema_id: str
    ) -> AnonCredsSchemaInfo:
        """Get a schema info from the registry."""
        return await super().get_schema_info_by_id(schema_id)



"""Ledger admin routes."""

import json
import logging

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields, validate

from ..admin.decorators.auth import tenant_authentication
from ..admin.request_context import AdminRequestContext
from ..connections.models.conn_record import ConnRecord
from ..messaging.models.base import BaseModelError
from ..messaging.models.openapi import OpenAPISchema
from ..messaging.valid import (
    ENDPOINT_EXAMPLE,
    ENDPOINT_TYPE_EXAMPLE,
    ENDPOINT_TYPE_VALIDATE,
    ENDPOINT_VALIDATE,
    INDY_DID_EXAMPLE,
    INDY_DID_VALIDATE,
    INT_EPOCH_EXAMPLE,
    INT_EPOCH_VALIDATE,
    RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
    RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
    UUID4_EXAMPLE,
)
from ..multitenant.base import BaseMultitenantManager
from ..protocols.endorse_transaction.v1_0.manager import (
    TransactionManager,
    TransactionManagerError,
)
from ..protocols.endorse_transaction.v1_0.models.transaction_record import (
    TransactionRecord,
    TransactionRecordSchema,
)
from ..protocols.endorse_transaction.v1_0.util import (
    get_endorser_connection_id,
    is_author_role,
)
from ..storage.error import StorageError, StorageNotFoundError
from ..wallet.error import WalletError, WalletNotFoundError
from .base import BaseLedger
from .base import Role as LedgerRole
from .endpoint_type import EndpointType
from .error import BadLedgerRequestError, LedgerError, LedgerTransactionError
from .multiple_ledger.base_manager import (
    BaseMultipleLedgerManager,
    MultipleLedgerManagerError,
)
from .multiple_ledger.ledger_config_schema import (
    ConfigurableWriteLedgersSchema,
    LedgerConfigListSchema,
    WriteLedgerSchema,
)
from .multiple_ledger.ledger_requests_executor import (
    GET_ENDPOINT_FOR_DID,
    GET_KEY_FOR_DID,
    GET_NYM_ROLE,
    IndyLedgerRequestsExecutor,
)
from .util import notify_register_did_event

LOGGER = logging.getLogger(__name__)


class LedgerModulesResultSchema(OpenAPISchema):
    """Schema for the modules endpoint."""


class AMLRecordSchema(OpenAPISchema):
    """Ledger AML record."""

    version = fields.Str()
    aml = fields.Dict(fields.Str(), fields.Str())
    amlContext = fields.Str()


class TAARecordSchema(OpenAPISchema):
    """Ledger TAA record."""

    version = fields.Str()
    text = fields.Str()
    digest = fields.Str()


class TAAAcceptanceSchema(OpenAPISchema):
    """TAA acceptance record."""

    mechanism = fields.Str()
    time = fields.Int(
        validate=INT_EPOCH_VALIDATE,
        metadata={"strict": True, "example": INT_EPOCH_EXAMPLE},
    )


class TAAInfoSchema(OpenAPISchema):
    """Transaction author agreement info."""

    aml_record = fields.Nested(AMLRecordSchema())
    taa_record = fields.Nested(TAARecordSchema())
    taa_required = fields.Bool()
    taa_accepted = fields.Nested(TAAAcceptanceSchema())


class TAAResultSchema(OpenAPISchema):
    """Result schema for a transaction author agreement."""

    result = fields.Nested(TAAInfoSchema())


class TAAAcceptSchema(OpenAPISchema):
    """Input schema for accepting the TAA."""

    version = fields.Str()
    text = fields.Str()
    mechanism = fields.Str()


class RegisterLedgerNymQueryStringSchema(OpenAPISchema):
    """Query string parameters and validators for register ledger nym request."""

    did = fields.Str(
        required=True,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "DID to register", "example": INDY_DID_EXAMPLE},
    )
    verkey = fields.Str(
        required=True,
        validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
        metadata={
            "description": "Verification key",
            "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
        },
    )
    alias = fields.Str(
        required=False, metadata={"description": "Alias", "example": "Barry"}
    )
    role = fields.Str(
        required=False,
        validate=validate.OneOf(
            [r.name for r in LedgerRole if isinstance(r.value[0], int)] + ["reset"]
        ),
        metadata={"description": "Role"},
    )


class CreateDidTxnForEndorserOptionSchema(OpenAPISchema):
    """Class for user to input whether to create a transaction for endorser or not."""

    create_transaction_for_endorser = fields.Boolean(
        required=False,
        metadata={"description": "Create Transaction For Endorser's signature"},
    )


class SchemaConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


class QueryStringDIDSchema(OpenAPISchema):
    """Parameters and validators for query string with DID only."""

    did = fields.Str(
        required=True,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "DID of interest", "example": INDY_DID_EXAMPLE},
    )


class QueryStringEndpointSchema(OpenAPISchema):
    """Parameters and validators for query string with DID and endpoint type."""

    did = fields.Str(
        required=True,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "DID of interest", "example": INDY_DID_EXAMPLE},
    )
    endpoint_type = fields.Str(
        required=False,
        validate=ENDPOINT_TYPE_VALIDATE,
        metadata={
            "description": (
                f"Endpoint type of interest (default '{EndpointType.ENDPOINT.w3c}')"
            ),
            "example": ENDPOINT_TYPE_EXAMPLE,
        },
    )


class TxnOrRegisterLedgerNymResponseSchema(OpenAPISchema):
    """Response schema for ledger nym registration."""

    success = fields.Bool(
        metadata={
            "description": "Success of nym registration operation",
            "example": True,
        }
    )

    txn = fields.Nested(
        TransactionRecordSchema(),
        required=False,
        metadata={"description": "DID transaction to endorse"},
    )


class GetNymRoleResponseSchema(OpenAPISchema):
    """Response schema to get nym role operation."""

    role = fields.Str(
        validate=validate.OneOf([r.name for r in LedgerRole]),
        metadata={"description": "Ledger role", "example": LedgerRole.ENDORSER.name},
    )


class GetDIDVerkeyResponseSchema(OpenAPISchema):
    """Response schema to get DID verkey."""

    verkey = fields.Str(
        allow_none=True,
        validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
        metadata={
            "description": "Full verification key",
            "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
        },
    )


class GetDIDEndpointResponseSchema(OpenAPISchema):
    """Response schema to get DID endpoint."""

    endpoint = fields.Str(
        allow_none=True,
        validate=ENDPOINT_VALIDATE,
        metadata={"description": "Full verification key", "example": ENDPOINT_EXAMPLE},
    )


class WriteLedgerRequestSchema(OpenAPISchema):
    """Schema for setting ledger_id for the write ledger."""

    ledger_id = fields.Str(required=True)


@docs(
    tags=["ledger"],
    summary="Send a NYM registration to the ledger.",
)
@querystring_schema(RegisterLedgerNymQueryStringSchema())
@querystring_schema(CreateDidTxnForEndorserOptionSchema())
@querystring_schema(SchemaConnIdMatchInfoSchema())
@response_schema(TxnOrRegisterLedgerNymResponseSchema(), 200, description="")
@tenant_authentication
async def register_ledger_nym(request: web.BaseRequest):
    """Request handler for registering a NYM with the ledger.

    Args:
        request: aiohttp request object
    """
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]
    async with context.profile.session() as session:
        ledger = session.inject_or(BaseLedger)
        if not ledger:
            reason = "No Indy ledger available"
            if not session.settings.get_value("wallet.type"):
                reason += ": missing wallet-type?"
            raise web.HTTPForbidden(reason=reason)

    did = request.query.get("did")
    verkey = request.query.get("verkey")
    if not did or not verkey:
        raise web.HTTPBadRequest(reason="Request query must include both did and verkey")

    alias = request.query.get("alias")
    role = request.query.get("role")
    if role == "reset":  # indy: empty to reset, null for regular user
        role = ""  # visually: confusing - correct 'reset' to empty string here

    create_transaction_for_endorser = json.loads(
        request.query.get("create_transaction_for_endorser", "false")
    )
    write_ledger = not create_transaction_for_endorser
    endorser_did = None
    connection_id = request.query.get("conn_id")

    # check if we need to endorse
    if is_author_role(context.profile):
        # authors cannot write to the ledger
        write_ledger = False
        create_transaction_for_endorser = True
        if not connection_id:
            # author has not provided a connection id, so determine which to use
            connection_id = await get_endorser_connection_id(context.profile)
            if not connection_id:
                raise web.HTTPBadRequest(reason="No endorser connection found")

    if not write_ledger:
        try:
            async with context.profile.session() as session:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, connection_id
                )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except BaseModelError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        async with context.profile.session() as session:
            endorser_info = await connection_record.metadata_get(session, "endorser_info")
        if not endorser_info:
            raise web.HTTPForbidden(
                reason=(
                    "Endorser Info is not set up in "
                    "connection metadata for this connection record"
                )
            )
        if "endorser_did" not in endorser_info.keys():
            raise web.HTTPForbidden(
                reason=(
                    ' "endorser_did" is not set in "endorser_info"'
                    " in connection metadata for this connection record"
                )
            )
        endorser_did = endorser_info["endorser_did"]

    meta_data = {"did": did, "verkey": verkey, "alias": alias, "role": role}
    success = False
    txn = None
    async with ledger:
        try:
            # if we are an author check if we have a public DID or not
            write_ledger_nym_transaction = True
            # special case - if we are an author with no public DID
            if create_transaction_for_endorser:
                public_info = await ledger.get_wallet_public_did()
                if not public_info:
                    write_ledger_nym_transaction = False
                    success = False
                    txn = {"signed_txn": json.dumps(meta_data)}
            if write_ledger_nym_transaction:
                (success, txn) = await ledger.register_nym(
                    did,
                    verkey,
                    alias,
                    role,
                    write_ledger=write_ledger,
                    endorser_did=endorser_did,
                )
        except LedgerTransactionError as err:
            raise web.HTTPForbidden(reason=err.roll_up)
        except LedgerError as err:
            raise web.HTTPBadRequest(reason=err.roll_up)
        except WalletNotFoundError as err:
            raise web.HTTPForbidden(reason=err.roll_up)
        except WalletError as err:
            raise web.HTTPBadRequest(
                reason=(
                    f"Registered NYM for DID {did} on ledger but could not "
                    f"replace metadata in wallet: {err.roll_up}"
                )
            )

    if not create_transaction_for_endorser:
        # Notify event
        await notify_register_did_event(context.profile, did, meta_data)
        return web.json_response({"success": success})
    else:
        transaction_mgr = TransactionManager(context.profile)
        try:
            transaction = await transaction_mgr.create_record(
                messages_attach=txn["signed_txn"],
                connection_id=connection_id,
                meta_data=meta_data,
            )
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        # if auto-request, send the request to the endorser
        if context.settings.get_value("endorser.auto_request"):
            try:
                endorser_write_txn = not write_ledger_nym_transaction
                transaction, transaction_request = await transaction_mgr.create_request(
                    transaction=transaction,
                    author_goal_code=(
                        TransactionRecord.REGISTER_PUBLIC_DID
                        if endorser_write_txn
                        else None
                    ),
                    signer_goal_code=(
                        TransactionRecord.WRITE_DID_TRANSACTION
                        if endorser_write_txn
                        else None
                    ),
                    # TODO see if we need to parametrize these params
                    # expires_time=expires_time,
                )
                txn = transaction.serialize()
            except (StorageError, TransactionManagerError) as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

            await outbound_handler(transaction_request, connection_id=connection_id)

        return web.json_response({"success": success, "txn": txn})


@docs(
    tags=["ledger"],
    summary="Get the role from the NYM registration of a public DID.",
)
@querystring_schema(QueryStringDIDSchema)
@response_schema(GetNymRoleResponseSchema(), 200, description="")
@tenant_authentication
async def get_nym_role(request: web.BaseRequest):
    """Request handler for getting the role from the NYM registration of a public DID.

    Args:
        request: aiohttp request object
    """
    context: AdminRequestContext = request["context"]

    did = request.query.get("did")
    if not did:
        raise web.HTTPBadRequest(reason="Request query must include DID")

    async with context.profile.session() as session:
        multitenant_mgr = session.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(context.profile)
        else:
            ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)
        ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
            did,
            txn_record_type=GET_NYM_ROLE,
        )
        if not ledger:
            reason = "No Indy ledger available"
            if not session.settings.get_value("wallet.type"):
                reason += ": missing wallet-type?"
            raise web.HTTPForbidden(reason=reason)

    async with ledger:
        try:
            role = await ledger.get_nym_role(did)
        except LedgerTransactionError as err:
            raise web.HTTPForbidden(reason=err.roll_up)
        except BadLedgerRequestError as err:
            raise web.HTTPNotFound(reason=err.roll_up)
        except LedgerError as err:
            raise web.HTTPBadRequest(reason=err.roll_up)

    if ledger_id:
        return web.json_response({"ledger_id": ledger_id, "role": role.name})
    else:
        return web.json_response({"role": role.name})


@docs(tags=["ledger"], summary="Rotate key pair for public DID.")
@response_schema(LedgerModulesResultSchema(), 200, description="")
@tenant_authentication
async def rotate_public_did_keypair(request: web.BaseRequest):
    """Request handler for rotating key pair associated with public DID.

    Args:
        request: aiohttp request object
    """
    context: AdminRequestContext = request["context"]
    async with context.profile.session() as session:
        ledger = session.inject_or(BaseLedger)
        if not ledger:
            reason = "No Indy ledger available"
            if not session.settings.get_value("wallet.type"):
                reason += ": missing wallet-type?"
            raise web.HTTPForbidden(reason=reason)
    async with ledger:
        try:
            await ledger.rotate_public_did_keypair()  # do not take seed over the wire
        except (WalletError, BadLedgerRequestError) as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({})


@docs(
    tags=["ledger"],
    summary="Get the verkey for a DID from the ledger.",
)
@querystring_schema(QueryStringDIDSchema())
@response_schema(GetDIDVerkeyResponseSchema(), 200, description="")
@tenant_authentication
async def get_did_verkey(request: web.BaseRequest):
    """Request handler for getting a verkey for a DID from the ledger.

    Args:
        request: aiohttp request object
    """
    context: AdminRequestContext = request["context"]

    did = request.query.get("did")
    if not did:
        raise web.HTTPBadRequest(reason="Request query must include DID")

    async with context.profile.session() as session:
        multitenant_mgr = session.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(context.profile)
        else:
            ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)
        ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
            did,
            txn_record_type=GET_KEY_FOR_DID,
        )
        if not ledger:
            reason = "No ledger available"
            if not session.settings.get_value("wallet.type"):
                reason += ": missing wallet-type?"
            raise web.HTTPForbidden(reason=reason)

    async with ledger:
        try:
            result = await ledger.get_key_for_did(did)
            if not result:
                raise web.HTTPNotFound(reason=f"DID {did} is not on the ledger")
        except LedgerError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    if ledger_id:
        return web.json_response({"ledger_id": ledger_id, "verkey": result})
    else:
        return web.json_response({"verkey": result})


@docs(
    tags=["ledger"],
    summary="Get the endpoint for a DID from the ledger.",
)
@querystring_schema(QueryStringEndpointSchema())
@response_schema(GetDIDEndpointResponseSchema(), 200, description="")
@tenant_authentication
async def get_did_endpoint(request: web.BaseRequest):
    """Request handler for getting a verkey for a DID from the ledger.

    Args:
        request: aiohttp request object
    """
    context: AdminRequestContext = request["context"]

    did = request.query.get("did")
    if not did:
        raise web.HTTPBadRequest(reason="Request query must include DID")

    async with context.profile.session() as session:
        multitenant_mgr = session.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(context.profile)
        else:
            ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)
        ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
            did,
            txn_record_type=GET_ENDPOINT_FOR_DID,
        )
        if not ledger:
            reason = "No Indy ledger available"
            if not session.settings.get_value("wallet.type"):
                reason += ": missing wallet-type?"
            raise web.HTTPForbidden(reason=reason)
    endpoint_type = EndpointType.get(
        request.query.get("endpoint_type", EndpointType.ENDPOINT.w3c)
    )

    async with ledger:
        try:
            r = await ledger.get_endpoint_for_did(did, endpoint_type)
        except LedgerError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    if ledger_id:
        return web.json_response({"ledger_id": ledger_id, "endpoint": r})
    else:
        return web.json_response({"endpoint": r})


@docs(tags=["ledger"], summary="Fetch the current transaction author agreement, if any")
@response_schema(TAAResultSchema, 200, description="")
@tenant_authentication
async def ledger_get_taa(request: web.BaseRequest):
    """Request handler for fetching the transaction author agreement.

    Args:
        request: aiohttp request object

    Returns:
        The TAA information including the AML

    """
    context: AdminRequestContext = request["context"]
    async with context.profile.session() as session:
        ledger = session.inject_or(BaseLedger)
        if not ledger:
            reason = "No Indy ledger available"
            if not session.settings.get_value("wallet.type"):
                reason += ": missing wallet-type?"
            raise web.HTTPForbidden(reason=reason)

    async with ledger:
        try:
            taa_info = await ledger.get_txn_author_agreement()
            accepted = None
            if taa_info["taa_required"]:
                accept_record = await ledger.get_latest_txn_author_acceptance()
                if accept_record:
                    accepted = {
                        "mechanism": accept_record["mechanism"],
                        "time": accept_record["time"],
                    }
            taa_info["taa_accepted"] = accepted
        except LedgerError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"result": taa_info})


@docs(tags=["ledger"], summary="Accept the transaction author agreement")
@request_schema(TAAAcceptSchema)
@response_schema(LedgerModulesResultSchema(), 200, description="")
@tenant_authentication
async def ledger_accept_taa(request: web.BaseRequest):
    """Request handler for accepting the current transaction author agreement.

    Args:
        request: aiohttp request object

    Returns:
        The DID list response

    """
    context: AdminRequestContext = request["context"]
    async with context.profile.session() as session:
        ledger = session.inject_or(BaseLedger)
        if not ledger:
            reason = "No Indy ledger available"
            if not session.settings.get_value("wallet.type"):
                reason += ": missing wallet-type?"
            raise web.HTTPForbidden(reason=reason)

    accept_input = await request.json()
    LOGGER.info(">>> accepting TAA with: %s", accept_input)
    async with ledger:
        try:
            taa_info = await ledger.get_txn_author_agreement()
            if not taa_info["taa_required"]:
                raise web.HTTPBadRequest(
                    reason=f"Ledger {ledger.pool_name} TAA not available"
                )
            LOGGER.info("TAA on ledger: ", taa_info)
            # this is a bit of a hack, but the "\ufeff" code is included in the
            # ledger TAA and digest calculation, so it needs to be included in the
            # TAA text that the user is accepting
            # (if you copy the TAA text using swagger it won't include this character)
            if taa_info["taa_record"]["text"].startswith("\ufeff"):
                if not accept_input["text"].startswith("\ufeff"):
                    LOGGER.info(
                        ">>> pre-pending -endian character to TAA acceptance text"
                    )
                    accept_input["text"] = "\ufeff" + accept_input["text"]
            taa_record = {
                "version": accept_input["version"],
                "text": accept_input["text"],
                "digest": ledger.taa_digest(
                    accept_input["version"],
                    accept_input["text"],
                ),
            }
            taa_record_digest = taa_record["digest"]
            LOGGER.info(">>> accepting with digest: %s", taa_record_digest)
            await ledger.accept_txn_author_agreement(
                taa_record, accept_input["mechanism"]
            )
        except (LedgerError, StorageError) as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({})


@docs(tags=["ledger"], summary="Fetch list of available write ledgers")
@response_schema(ConfigurableWriteLedgersSchema, 200, description="")
@tenant_authentication
async def get_write_ledgers(request: web.BaseRequest):
    """Request handler for fetching the list of available write ledgers.

    Args:
        request: aiohttp request object

    Returns:
        The list of write ledgers

    """
    context: AdminRequestContext = request["context"]
    async with context.profile.session() as session:
        multiledger_mgr = session.inject_or(BaseMultipleLedgerManager)
    if not multiledger_mgr:
        return web.json_response({"write_ledgers": ["default"]})
    available_write_ledgers = await multiledger_mgr.get_write_ledgers()
    return web.json_response({"write_ledgers": available_write_ledgers})


@docs(tags=["ledger"], summary="Fetch the current write ledger")
@response_schema(WriteLedgerSchema, 200, description="")
@tenant_authentication
async def get_write_ledger(request: web.BaseRequest):
    """Request handler for fetching the currently set write ledger.

    Args:
        request: aiohttp request object

    Returns:
        The write ledger identifier

    """
    context: AdminRequestContext = request["context"]
    async with context.profile.session() as session:
        multiledger_mgr = session.inject_or(BaseMultipleLedgerManager)
        write_ledger = session.inject(BaseLedger)
    if not multiledger_mgr:
        return web.json_response({"ledger_id": "default"})
    ledger_id = await multiledger_mgr.get_ledger_id_by_ledger_pool_name(
        write_ledger.pool_name
    )
    return web.json_response({"ledger_id": ledger_id})


@docs(tags=["ledger"], summary="Set write ledger")
@match_info_schema(WriteLedgerRequestSchema())
@response_schema(WriteLedgerSchema, 200, description="")
@tenant_authentication
async def set_write_ledger(request: web.BaseRequest):
    """Request handler for setting write ledger.

    Args:
        request: aiohttp request object

    Returns:
        The set write ledger identifier

    """
    context: AdminRequestContext = request["context"]
    async with context.profile.session() as session:
        multiledger_mgr = session.inject_or(BaseMultipleLedgerManager)
    if not multiledger_mgr:
        return web.json_response({"write_ledger": "default"})
    req_ledger_id = request.match_info.get("ledger_id")
    try:
        set_ledger_id = await multiledger_mgr.set_profile_write_ledger(
            ledger_id=req_ledger_id,
            profile=context.profile,
        )
    except MultipleLedgerManagerError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    return web.json_response({"write_ledger": set_ledger_id})


@docs(tags=["ledger"], summary="Fetch the multiple ledger configuration currently in use")
@response_schema(LedgerConfigListSchema, 200, description="")
@tenant_authentication
async def get_ledger_config(request: web.BaseRequest):
    """Request handler for fetching the ledger configuration list in use.

    Args:
        request: aiohttp request object

    Returns:
        Ledger configuration list

    """
    context: AdminRequestContext = request["context"]
    async with context.profile.session() as session:
        multiledger_mgr = session.inject_or(BaseMultipleLedgerManager)
        if not multiledger_mgr:
            reason = "Multiple ledger support not enabled"
            raise web.HTTPForbidden(reason=reason)
        ledger_config_list = session.settings.get_value("ledger.ledger_config_list")
        config_ledger_dict = {"production_ledgers": [], "non_production_ledgers": []}
        production_ledger_keys = (await multiledger_mgr.get_prod_ledgers()).keys()
        non_production_ledger_keys = (await multiledger_mgr.get_nonprod_ledgers()).keys()
        config_ledger_ids_set = set()
        for config in ledger_config_list:
            ledger_id = config.get("id")
            config_ledger_ids_set.add(ledger_id)
            # removing genesis_transactions
            config = {
                key: val for key, val in config.items() if key != "genesis_transactions"
            }
            if ledger_id in production_ledger_keys:
                config_ledger_dict.get("production_ledgers").append(config)
            if ledger_id in non_production_ledger_keys:
                config_ledger_dict.get("non_production_ledgers").append(config)
        diff_prod_ledger_ids_set = set(production_ledger_keys) - config_ledger_ids_set
        for diff_prod_ledger_id in diff_prod_ledger_ids_set:
            config_ledger_dict.get("production_ledgers").append(
                {
                    "id": diff_prod_ledger_id,
                    "desc": "ledger configured outside --genesis-transactions-list",
                }
            )
    return web.json_response(config_ledger_dict)


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.post("/ledger/register-nym", register_ledger_nym),
            web.get("/ledger/get-nym-role", get_nym_role, allow_head=False),
            web.patch("/ledger/rotate-public-did-keypair", rotate_public_did_keypair),
            web.get("/ledger/did-verkey", get_did_verkey, allow_head=False),
            web.get("/ledger/did-endpoint", get_did_endpoint, allow_head=False),
            web.get("/ledger/taa", ledger_get_taa, allow_head=False),
            web.post("/ledger/taa/accept", ledger_accept_taa),
            web.get("/ledger/get-write-ledger", get_write_ledger, allow_head=False),
            web.put("/ledger/{ledger_id}/set-write-ledger", set_write_ledger),
            web.get(
                "/ledger/get-write-ledgers",
                get_write_ledgers,
                allow_head=False,
            ),
            web.get("/ledger/config", get_ledger_config, allow_head=False),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "ledger",
            "description": "Interaction with ledger",
            "externalDocs": {
                "description": "Overview",
                "url": (
                    "https://hyperledger-indy.readthedocs.io/projects/plenum/"
                    "en/latest/storage.html#ledger"
                ),
            },
        }
    )



"""Represents a feature discovery queries message."""

from typing import Sequence

from marshmallow import EXCLUDE, fields, validate

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.models.base import BaseModel, BaseModelSchema
from ..message_types import PROTOCOL_PACKAGE, QUERIES

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.queries_handler.QueriesHandler"


class QueryItem(BaseModel):
    """Defines QueryItem field."""

    class Meta:
        """QueryItem metadata."""

        schema_class = "QueryItemSchema"

    def __init__(
        self,
        *,
        feature_type: str,
        match: str,
    ):
        """Initialize QueryItem."""
        self.feature_type = feature_type
        self.match = match


class QueryItemSchema(BaseModelSchema):
    """Single QueryItem Schema."""

    class Meta:
        """QueryItemSchema metadata."""

        model_class = QueryItem
        unknown = EXCLUDE

    feature_type = fields.Str(
        required=True,
        data_key="feature-type",
        validate=validate.OneOf(["protocol", "goal-code"]),
        metadata={"description": "feature type"},
    )
    match = fields.Str(required=True, metadata={"description": "match"})


class Queries(AgentMessage):
    """Represents a discover-features v2 queries message.

    Used for inspecting what message types are supported by the agent.
    """

    class Meta:
        """Queries metadata."""

        handler_class = HANDLER_CLASS
        message_type = QUERIES
        schema_class = "QueriesSchema"

    def __init__(self, *, queries: Sequence[QueryItem] = None, **kwargs):
        """Initialize query message object.

        Args:
            queries: The query string to match against supported message types
            kwargs: Additional key word arguments for the message
        """
        super().__init__(**kwargs)
        self.queries = queries


class QueriesSchema(AgentMessageSchema):
    """Query message schema used in serialization/deserialization."""

    class Meta:
        """QuerySchema metadata."""

        model_class = Queries
        unknown = EXCLUDE

    queries = fields.List(fields.Nested(QueryItemSchema))



from unittest import IsolatedAsyncioTestCase, mock

from .....core.event_bus import EventBus, MockEventBus
from .....messaging.request_context import RequestContext
from .....utils.testing import create_test_profile
from .. import driver_service as test_module


class TestActionMenuService(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.context = RequestContext(self.profile)

    async def test_get_active_menu(self):
        mock_event_bus = MockEventBus()
        self.context.profile.context.injector.bind_instance(EventBus, mock_event_bus)

        self.menu_service = await test_module.DriverMenuService.service_handler()(
            self.context
        )

        connection = mock.MagicMock()
        connection.connection_id = "connid"
        thread_id = "thid"

        await self.menu_service.get_active_menu(
            self.context.profile, connection, thread_id
        )

        assert len(mock_event_bus.events) == 1
        (_, event) = mock_event_bus.events[0]
        assert event.topic == "acapy::actionmenu::get-active-menu"
        assert event.payload == {
            "connection_id": connection.connection_id,
            "thread_id": thread_id,
        }

    async def test_perform_menu_action(self):
        mock_event_bus = MockEventBus()
        self.context.profile.context.injector.bind_instance(EventBus, mock_event_bus)

        self.menu_service = await test_module.DriverMenuService.service_handler()(
            self.context
        )

        action_name = "action"
        action_params = {"a": 1, "b": 2}
        connection = mock.MagicMock()
        connection.connection_id = "connid"
        thread_id = "thid"

        await self.menu_service.perform_menu_action(
            self.context.profile,
            action_name,
            action_params,
            connection,
            thread_id,
        )

        assert len(mock_event_bus.events) == 1
        (_, event) = mock_event_bus.events[0]
        assert event.topic == "acapy::actionmenu::perform-menu-action"
        assert event.payload == {
            "connection_id": connection.connection_id,
            "thread_id": thread_id,
            "action_name": action_name,
            "action_params": action_params,
        }



"""HTTP utility methods."""

import asyncio
import logging
import urllib.parse
from typing import Optional

from aiohttp import BaseConnector, ClientError, ClientResponse, ClientSession, FormData
from aiohttp.web import HTTPConflict

from ..core.error import BaseError
from .repeat import RepeatSequence

LOGGER = logging.getLogger(__name__)


class FetchError(BaseError):
    """Error raised when an HTTP fetch fails."""


class PutError(BaseError):
    """Error raised when an HTTP put fails."""


async def fetch_stream(
    url: str,
    *,
    headers: Optional[dict] = None,
    retry: bool = True,
    max_attempts: int = 5,
    interval: float = 1.0,
    backoff: float = 0.25,
    request_timeout: float = 10.0,
    connector: Optional[BaseConnector] = None,
    session: Optional[ClientSession] = None,
):
    """Fetch from an HTTP server with automatic retries and timeouts.

    Args:
        url: the address to fetch
        headers: an optional dict of headers to send
        retry: flag to retry the fetch
        max_attempts: the maximum number of attempts to make
        interval: the interval between retries, in seconds
        backoff: the backoff interval, in seconds
        request_timeout: the HTTP request timeout, in seconds
        connector: an optional existing BaseConnector
        session: a shared ClientSession
        json: flag to parse the result as JSON

    """
    limit = max_attempts if retry else 1
    if not session:
        session = ClientSession(
            connector=connector, connector_owner=(not connector), trust_env=True
        )
    async with session:
        async for attempt in RepeatSequence(limit, interval, backoff):
            try:
                async with attempt.timeout(request_timeout):
                    response: ClientResponse = await session.get(url, headers=headers)
                    if response.status < 200 or response.status >= 300:
                        raise ClientError(
                            f"Bad response from server: {response.status} - "
                            f"{response.reason}"
                        )
                    return response.content
            except (ClientError, asyncio.TimeoutError) as e:
                if attempt.final:
                    raise FetchError("Exceeded maximum fetch attempts") from e


async def fetch(
    url: str,
    *,
    headers: Optional[dict] = None,
    retry: bool = True,
    max_attempts: int = 5,
    interval: float = 1.0,
    backoff: float = 0.25,
    request_timeout: float = 10.0,
    connector: Optional[BaseConnector] = None,
    session: Optional[ClientSession] = None,
    json: bool = False,
):
    """Fetch from an HTTP server with automatic retries and timeouts.

    Args:
        url: the address to fetch
        headers: an optional dict of headers to send
        retry: flag to retry the fetch
        max_attempts: the maximum number of attempts to make
        interval: the interval between retries, in seconds
        backoff: the backoff interval, in seconds
        request_timeout: the HTTP request timeout, in seconds
        connector: an optional existing BaseConnector
        session: a shared ClientSession
        json: flag to parse the result as JSON

    """
    limit = max_attempts if retry else 1
    if not session:
        session = ClientSession(
            connector=connector, connector_owner=(not connector), trust_env=True
        )
    async with session:
        async for attempt in RepeatSequence(limit, interval, backoff):
            try:
                async with attempt.timeout(request_timeout):
                    response: ClientResponse = await session.get(url, headers=headers)
                    if response.status < 200 or response.status >= 300:
                        raise ClientError(
                            f"Bad response from server: {response.status} - "
                            f"{response.reason}"
                        )
                    return await (response.json() if json else response.text())
            except (ClientError, asyncio.TimeoutError) as e:
                if attempt.final:
                    raise FetchError("Exceeded maximum fetch attempts") from e


async def put_file(
    url: str,
    file_data: dict,
    extra_data: dict,
    *,
    retry: bool = True,
    max_attempts: int = 5,
    interval: float = 1.0,
    backoff: float = 0.25,
    request_timeout: float = 10.0,
    connector: Optional[BaseConnector] = None,
    session: Optional[ClientSession] = None,
    json: bool = False,
):
    """Put to HTTP server with automatic retries and timeouts.

    Args:
        url: the address to use
        file_data: dict with data key and path of file to upload
        extra_data: further content to include in data to put
        headers: an optional dict of headers to send
        retry: flag to retry the fetch
        max_attempts: the maximum number of attempts to make
        interval: the interval between retries, in seconds
        backoff: the backoff interval, in seconds
        request_timeout: the HTTP request timeout, in seconds
        connector: an optional existing BaseConnector
        session: a shared ClientSession
        json: flag to parse the result as JSON

    """
    (data_key, file_path) = list(file_data.items())[0]
    limit = max_attempts if retry else 1

    if not session:
        session = ClientSession(
            connector=connector, connector_owner=(not connector), trust_env=True
        )
    async with session:
        async for attempt in RepeatSequence(limit, interval, backoff):
            try:
                async with attempt.timeout(request_timeout):
                    formdata = FormData()
                    try:
                        fp = open(file_path, "rb")
                    except OSError as e:
                        raise PutError("Error opening file for upload") from e
                    if extra_data:
                        for k, v in extra_data.items():
                            formdata.add_field(k, v)
                    formdata.add_field(
                        data_key, fp, content_type="application/octet-stream"
                    )
                    response: ClientResponse = await session.put(
                        url, data=formdata, allow_redirects=False
                    )
                    if (
                        # redirect codes
                        response.status in (301, 302, 303, 307, 308) and not attempt.final
                    ):
                        # NOTE: a redirect counts as another upload attempt
                        to_url = response.headers.get("Location")
                        if not to_url:
                            raise PutError("Redirect missing target URL")
                        try:
                            parsed_to = urllib.parse.urlsplit(to_url)
                            parsed_from = urllib.parse.urlsplit(url)
                        except ValueError:
                            raise PutError("Invalid redirect URL")
                        if parsed_to.hostname != parsed_from.hostname:
                            raise PutError("Redirect denied: hostname mismatch")
                        url = to_url
                        LOGGER.info("Upload redirect: %s", to_url)
                    elif (response.status < 200 or response.status >= 300) and (
                        response.status != HTTPConflict.status_code
                    ):
                        raise ClientError(
                            f"Bad response from server: {response.status}, "
                            f"{response.reason}"
                        )
                    else:
                        return await (response.json() if json else response.text())
            except (ClientError, asyncio.TimeoutError) as e:
                if isinstance(e, ClientError):
                    LOGGER.warning("Upload error: %s", e)
                else:
                    LOGGER.warning("Upload error: request timed out")
                if attempt.final:
                    raise PutError("Exceeded maximum upload attempts") from e



from unittest import IsolatedAsyncioTestCase, mock

from .. import stats as test_module


class TestStatsTracer(IsolatedAsyncioTestCase):
    def setUp(self):
        self.context = mock.MagicMock(
            socket_timer=mock.MagicMock(
                stop=mock.MagicMock(side_effect=AttributeError("wrong"))
            )
        )
        self.tracer = test_module.StatsTracer(test_module.Collector(), "test")

    async def test_queued_start_stop(self):
        await self.tracer.connection_queued_start(None, self.context, None)
        await self.tracer.connection_queued_end(None, self.context, None)

    async def test_connection_ready_error_pass(self):
        await self.tracer.connection_ready(None, self.context, None)



"""EddsaJcs2022 cryptosuite."""

from datetime import datetime
from hashlib import sha256

import canonicaljson

from ....core.error import BaseError
from ....core.profile import ProfileSession
from ....utils.multiformats import multibase
from ....wallet.base import BaseWallet
from ....wallet.keys.manager import (
    MultikeyManager,
    key_type_from_multikey,
    multikey_to_verkey,
)
from ..errors import PROBLEM_DETAILS
from ..models.options import DataIntegrityProofOptions
from ..models.proof import DataIntegrityProof
from ..models.verification_response import DataIntegrityVerificationResult, ProblemDetails


class CryptosuiteError(BaseError):
    """Generic Cryptosuite Error."""


class EddsaJcs2022:
    """EddsaJcs2022 cryptosuite.

    https://www.w3.org/TR/vc-di-eddsa/#eddsa-jcs-2022.
    """

    def __init__(self, *, session: ProfileSession):
        """Create new EddsaJcs2022 Cryptosuite instance.

        Args:
            session: ProfileSession to use during crypto operations.

        """
        super().__init__()
        self.session = session
        self.wallet = session.inject(BaseWallet)
        self.key_manager = MultikeyManager(session)

    async def create_proof(
        self, unsecured_data_document: dict, options: DataIntegrityProofOptions
    ):
        """Create proof algorithm.

        https://www.w3.org/TR/vc-di-eddsa/#create-proof-eddsa-jcs-2022.
        """
        proof = DataIntegrityProof.deserialize(options.serialize().copy())

        # Spec says to copy document context to the proof but it's unecessary IMO,
        # commenting out for the time being...

        # if '@context' in unsecured_data_document:
        #     proof['@context'] = unsecured_data_document['@context']

        proof_config = self.proof_configuration(proof)
        transformed_data = self.transformation(unsecured_data_document, options)
        hash_data = self.hashing(transformed_data, proof_config)
        proof_bytes = await self.proof_serialization(hash_data, options)

        proof.proof_value = multibase.encode(proof_bytes, "base58btc")

        return proof

    def proof_configuration(self, options: DataIntegrityProofOptions):
        """Proof configuration algorithm.

        https://www.w3.org/TR/vc-di-eddsa/#proof-configuration-eddsa-jcs-2022.
        """
        proof_config = options
        assert proof_config.type == "DataIntegrityProof", (
            'Expected proof.type to be "DataIntegrityProof'
        )
        assert proof_config.cryptosuite == "eddsa-jcs-2022", (
            'Expected proof.cryptosuite to be "eddsa-jcs-2022'
        )

        if proof_config.created:
            assert datetime.fromisoformat(proof_config.created)

        if proof_config.expires:
            assert datetime.fromisoformat(proof_config.expires)

        return self._canonicalize(proof_config.serialize())

    def transformation(
        self, unsecured_document: dict, options: DataIntegrityProofOptions
    ):
        """Transformation algorithm.

        https://www.w3.org/TR/vc-di-eddsa/#transformation-eddsa-jcs-2022.
        """
        assert options.type == "DataIntegrityProof", (
            "Expected proof.type to be `DataIntegrityProof`"
        )
        assert options.cryptosuite == "eddsa-jcs-2022", (
            "Expected proof.cryptosuite to be `eddsa-jcs-2022`"
        )

        return self._canonicalize(unsecured_document)

    def hashing(self, transformed_document: bytes, canonical_proof_config: bytes):
        """Hashing algorithm.

        https://www.w3.org/TR/vc-di-eddsa/#hashing-eddsa-jcs-2022.
        """
        return (
            sha256(canonical_proof_config).digest()
            + sha256(transformed_document).digest()
        )

    async def proof_serialization(
        self, hash_data: bytes, options: DataIntegrityProofOptions
    ):
        """Proof Serialization Algorithm.

        https://www.w3.org/TR/vc-di-eddsa/#proof-serialization-eddsa-jcs-2022.
        """
        # TODO encapsulate in a key manager method
        if options.verification_method.startswith("did:key:"):
            multikey = options.verification_method.split("#")[-1]
            key_info = await self.key_manager.from_multikey(multikey)

        else:
            key_info = await self.key_manager.from_kid(options.verification_method)

        return await self.wallet.sign_message(
            message=hash_data,
            from_verkey=multikey_to_verkey(key_info["multikey"]),
        )

    def _canonicalize(self, data: dict):
        """Json canonicalization."""
        return canonicaljson.encode_canonical_json(data)

    async def verify_proof(self, secured_document: dict):
        """Verify proof algorithm.

        https://www.w3.org/TR/vc-di-eddsa/#verify-proof-eddsa-jcs-2022.
        """
        unsecured_document = secured_document.copy()
        proof = unsecured_document.pop("proof")
        proof_options = proof.copy()
        proof_bytes = multibase.decode(proof_options.pop("proofValue"))

        try:
            # Currently leaving context processing out of scope,
            # leaving code commented as it's technically an algorithm step.
            # Due to the cryptosuite being based on JSON canonicalization,
            # the integrity of the document is protected without RDF processing.

            # https://www.w3.org/TR/vc-data-integrity/#validating-contexts

            # assert secured_document['@context'] == proof_options['@context']
            # unsecured_document['@context'] = proof_options['@context']

            proof_options = DataIntegrityProofOptions.deserialize(proof_options)
            transformed_data = self.transformation(unsecured_document, proof_options)
            proof_config = self.proof_configuration(proof_options)
            hash_data = self.hashing(transformed_data, proof_config)
            verified = await self.proof_verification(
                hash_data, proof_bytes, proof_options
            )
            if not verified:
                raise CryptosuiteError("Invalid signature.")

        except CryptosuiteError as err:
            problem_detail = ProblemDetails.deserialize(
                PROBLEM_DETAILS["PROOF_VERIFICATION_ERROR"]
            )
            problem_detail.detail = str(err)
            return DataIntegrityVerificationResult(
                verified=False,
                proof=DataIntegrityProof.deserialize(proof),
                problem_details=[problem_detail],
            )

        return DataIntegrityVerificationResult(
            verified=True,
            proof=DataIntegrityProof.deserialize(proof),
            problem_details=[],
        )

    async def proof_verification(
        self, hash_data: bytes, proof_bytes: bytes, options: DataIntegrityProofOptions
    ):
        """Proof verification algorithm.

        https://www.w3.org/TR/vc-di-eddsa/#proof-verification-eddsa-jcs-2022.
        """
        multikey = await MultikeyManager(
            self.session
        ).resolve_multikey_from_verification_method_id(options.verification_method)
        verkey = multikey_to_verkey(multikey)
        key_type = key_type_from_multikey(multikey)
        return await self.wallet.verify_message(
            message=hash_data,
            signature=proof_bytes,
            from_verkey=verkey,
            key_type=key_type,
        )
